import numpy as np
from dataclasses import dataclass
from typing import Union, Tuple
import itertools as itert
import scipy.optimize as scop
import copy
import forsys.virtual_edges as ve
import forsys.borders as borders
from forsys.exceptions import BigEdgesBadlyCreated
import warnings
@dataclass
class ForceMatrix:
    """
    Class to build and solve the force matrix

    :param frame: Frame object from which matrix system is created
    :type frame: object
    :param externals_to_use: List of external vertices to use, or method to determine them. 
    :type externals_to_use: str or list
    :param term: If externals are incorporated, a term can be added to account for the third \
        force in the balance. area, perimeter, area-perimeter and ext are possible.
    :type term: str
    :param metadata: Extra parameters required by functions in the class. 
    :type metadata: dict
    :param timeseries: Time series dictionary of frames
    :type timeseries: dict
    """
    # TODO: Incorporate stress solution to the GeneralMatrix paradigm
    frame: object
    externals_to_use: Union[str, list]
    term: str
    metadata: dict
    timeseries: dict
    angle_limit: float = np.inf
    circle_fit_method: str = "dlite"

    def __post_init__(self):
        """Constructor method
        """
        self.map_vid_to_row = {}
        self.map_edge_to_column = {}
        self.deletes = set()

        if type(self.externals_to_use) == str:
            if self.externals_to_use == 'all':
                # self.externals_to_use = list(np.array(ve.get_border_edge(self.frame.big_edges_list, 
                #                                                         self.frame.vertices)).flatten())
                raise(NotImplementedError)
            elif self.externals_to_use == 'ext':
                self.externals_to_use = ve.get_border_from_angles_new(self.frame.big_edges_list, 
                                                                        self.frame.vertices)
                self.big_edges_to_use = self.frame.big_edges_list
            else:
                # Only solve on the triple junctions
                self.externals_to_use = []
                # TODO: make it work with the general matrix solver using self.frame.internal_big_edges
                # self.big_edges_to_use = self.frame.internal_big_edges_vertices
                self.big_edges_to_use, _ = self.get_angle_limited_edges()
        else:
            raise(NotImplementedError)

        tj_vertices = set()
        for big_edge in self.big_edges_to_use:
            tj_vertices.add(big_edge[0])
            tj_vertices.add(big_edge[-1])

        self.tj_vertices = list(tj_vertices)
        self.matrix = self._build_matrix()
        self.rhs = None
        self.velocity_matrix = None


    def _build_matrix(self) -> np.ndarray:
        """Build the stress inference matrix row by row

        :return: Stress inference matrix before least squares is applied
        :rtype: Matrix
        """
        max_rows = len(self.tj_vertices) * 2
        cols = len(self.big_edges_to_use) + len(self.externals_to_use) * 2
        mat = np.empty(shape=(max_rows, cols))
        position_index = 0
        for vid in self.tj_vertices:
            row_x, row_y = self.get_row(vid)
            non_zero_x = np.count_nonzero(row_x)
            non_zero_y = np.count_nonzero(row_y)
            at_least_three = non_zero_x >= 3 or non_zero_y >= 3
            less_than_four = non_zero_x < 4 and non_zero_y < 4
            less_than_four_condition = less_than_four if self.metadata.get("ignore_four", False) else True

            if at_least_three and less_than_four_condition:
                self.map_vid_to_row[vid] = position_index
                mat[position_index] = row_x
                mat[position_index + 1] = row_y
                position_index += 2

        # return matrix with populated rows
        return mat[:position_index]

    def get_row(self, vid: int) -> Tuple:
        """Create the general row by appending inference rows to external terms

        :param vid: ID of the vertex for the required row
        :type vid: int
        :return: The rows corresponding to the x and y components.
        :rtype: _type_
        """
        # create the two rows of the A matrix (row_x, row_y)
        arrx, arry = self.get_vertex_equation(vid)
        arrxF, arryF = self.get_external_term(vid)
        return np.concatenate((arrx, arrxF)), np.concatenate((arry, arryF))
        
    def get_external_term(self, vid: int) -> Tuple:
        """Get the external forces corresponding to the giving vertex in row form to append
        to the stress inference matrix

        :param vid: Current vertex ID
        :type vid: int
        :return: X and Y component of the external forces for the required vertex
        :rtype: Tuple
        """
        arrxF = np.zeros(len(self.externals_to_use)*2)
        arryF = np.zeros(len(self.externals_to_use)*2)
        if vid in self.externals_to_use:
            if self.term == 'area':
                fxVersor, fyVersor = borders.get_versors_area(self.frame.vertices, 
                                                                self.frame.edges, 
                                                                self.frame.cells, 
                                                                vid, 
                                                                self.metadata)
                # print("vID:", vid, " force : ", fxVersor, fyVersor)
            elif self.term == 'perimeter':
                fxVersor, fyVersor = borders.get_versors_perimeter(self.frame.vertices, 
                                                                    self.frame.edges, 
                                                                    self.frame.cells, 
                                                                    vid, 
                                                                    self.metadata)
            elif self.term == 'area-perimeter':
                axVersor, ayVersor = borders.get_versors_area(self.frame.vertices, 
                                                                self.frame.edges, 
                                                                self.frame.cells,
                                                                vid, 
                                                                self.metadata)
                pxVersor, pyVersor = borders.get_versors_perimeter(self.frame.vertices, 
                                                                    self.frame.edges, 
                                                                    self.frame.cells, 
                                                                    vid, 
                                                                    self.metadata)
                fxVersor = axVersor + pxVersor
                fyVersor = ayVersor + pyVersor
            elif self.term == 'none':
                fxVersor = 0
                fyVersor = 0
            else: # term='ext' is the default, also works for term='timeseries-velocity'
                fxVersor = 0
                fyVersor = 0
                for el in ve.get_versors(self.frame.vertices, self.frame.edges, vid):
                    fxVersor -= round(el[0], 3)
                    fyVersor -= round(el[1], 3)
            current_border_id = np.where(np.array(self.externals_to_use).astype(int) == vid)[0]

            arrxF[current_border_id*2] = round(fxVersor, 3)
            arryF[current_border_id * 2 + 1] = round(fyVersor, 3)
        return arrxF, arryF

    def get_vertex_equation(self, vid: int) -> Tuple:
        """Generate the stress inference row corresponding to the given vertex ID 

        :param vid: ID of the required vertex's equation
        :type vid: int
        :return: X and Y componenet of the stress inference matrix for the required vertex
        :rtype: Tuple
        """
        arrx = np.zeros(len(self.big_edges_to_use))
        arry = np.zeros(len(self.big_edges_to_use))
        vertex = self.frame.vertices[vid]
        vertex_big_edges = [self.frame.big_edges[beid] for beid in vertex.own_big_edges]
        vertex_big_edges_versors = [big_edge.get_versor_from_vertex(vid, fit_method=self.circle_fit_method) for big_edge in vertex_big_edges]
        # Find the three angles
        # TODO: Should something be done for 4-fold junctions ?
        # if len(vertex_big_edges) == 3:
        #     combinations = itert.combinations(vertex_big_edges_versors, r=2)
        #     angles = [np.arccos(np.dot(*combination)) for combination in combinations]
        #     if np.max(angles) >= self.angle_limit:
        #         self.deletes.add(vid)
        #         vertex_big_edges_versors = np.zeros((3, 2))

        for index, big_edge in enumerate(vertex_big_edges):
            if not big_edge.external and len(vertex.ownCells) > 2:
                try:
                    pos = ve.eid_from_vertex(self.big_edges_to_use, big_edge.get_vertices_ids())
                    versor = vertex_big_edges_versors[index]
                    arrx[pos] = versor[0]
                    arry[pos] = versor[1]
                except BigEdgesBadlyCreated:
                    continue

        return arrx, arry
    
    def get_angle_limited_edges(self):
        big_edges_vertices = [big_edge.get_vertices_ids() for big_edge in self.frame.internal_big_edges]
        tj_vertices = set([big_edge[0] for big_edge in big_edges_vertices])
        last = set([big_edge[-1] for big_edge in big_edges_vertices])
        tj_vertices.update(last)
        for vid in tj_vertices:
            vertex = self.frame.vertices[vid]
            vertex_big_edges = [self.frame.big_edges[beid] for beid in vertex.own_big_edges]
            vertex_big_edges_versors = [big_edge.get_versor_from_vertex(vid, fit_method=self.circle_fit_method) for big_edge in vertex_big_edges]
            # Find the three angles
            # TODO: Should something be done for 4-fold junctions ?
            # if len(vertex_big_edges) == 3:
            combinations = itert.combinations(vertex_big_edges_versors, r=2)
            angles = [np.arccos(np.dot(*combination)) for combination in combinations]
            if np.max(angles) >= self.angle_limit:
                self.deletes.add(vid)
        
        big_edges_to_use = copy.copy(self.frame.internal_big_edges_vertices)
        
        for _, big_edge in enumerate(big_edges_vertices):
            if (big_edge[0] in self.deletes) and (big_edge[-1] in self.deletes):
                big_edges_to_use.remove(big_edge)

        return big_edges_to_use, self.deletes


    def solve(self, timeseries: Union[str, list] = None, **kwargs) -> dict:
        """Solve the system of equations. The stress inference matrix must be built beforehand.

        :param timeseries: If a dynamical inference is required, the timeseries must be provided, defaults to None
        :type timeseries: Union[str, list], optional
        :return: Dictionary of edges and inferred stress. External forces are included if they were required
        :rtype: dict
        """
        np.seterr(all='raise')
        tote = len(self.big_edges_to_use)

        b, average_velocity = self.set_velocity_matrix(timeseries, **kwargs)

        solver_method = kwargs.get("method", None)
        if solver_method == "fix_stress":
            mprime, b, removed_index = self.fix_one_stress(b)
        elif solver_method == "lsq_linear":
            mprime, b = self.add_mean_one_before(b)
            removed_index = None
        elif solver_method == "lsq":
            mprime, b = self.add_mean_one(b)
            removed_index = None
        else:
            mprime, b = self.add_mean_one(b)
            removed_index = None

        b = b.astype(np.float64)

        self.rhs = b.T.round(4)

        mprime = mprime.astype(np.float64)
        # flatten b to convert it to a vector. Rounding is to keep old behavior (not sure if it's useful)
        b = b.astype(np.float64).flatten().round(3)
        try:
            if solver_method == "lsq_linear":
                solutions = scop.lsq_linear(mprime,
                                            b,
                                            bounds=(0.0, np.inf))
                xres = solutions["x"]
            elif solver_method == "lsq":
                try:
                    import lmfit as lmf
                except ModuleNotFoundError:
                    raise ModuleNotFoundError(f'lmfit is required for {solver_method=}')
                arguments = (mprime, b)
                x0_original = kwargs.get("initial_condition", np.ones(len(self.frame.internal_big_edges)))
                x0, removed_indices = self.get_new_initial_condition(x0_original, what="other")
                x0 = [val for val in x0 if val > 0]
                x0.append(1)

                def lmfit_cost(params, A, b):
                    _x = [params[name].value for name in params]
                    a_x = np.dot(np.array(A), np.array(_x)) 
                    vectorial_differences = a_x - b
                    return vectorial_differences

                def lmfit_cost_std(params, A, b):
                    _x = np.array([params[name].value for name in params])
                    a_x = np.dot(A, _x)
                    vectorial_differences = a_x - b
                    return vectorial_differences * (1 + 0.5 * _x.std())

                parameters = lmf.Parameters()
                for index, val in enumerate(x0):
                    naming = f"_{index}"
                    parameters.add(naming, val)
                    parameters[naming].min = 0

                if kwargs.get("use_std", False):
                    solution = lmf.minimize(lmfit_cost_std,
                                            params=parameters,
                                            args=arguments)
                else:
                    solution = lmf.minimize(lmfit_cost,
                                            params=parameters,
                                            args=arguments)
                # TODO: replace Matrix by ndarray in this code
                xres = [solution.params[name].value for name in solution.params]

                # reinsert all the removed spaces
                for index in removed_indices:
                    xres = xres.insert(index, -1)
            else:
                try:
                    xres = np.linalg.inv(mprime) @ b
                except np.linalg.LinAlgError:
                    raise ValueError("Singular matrix")
                
                if np.any([x < 0 for x in xres[:-1]]) and not kwargs.get("allow_negatives", True):
                    raise ValueError("Negative values detected")
        except (ValueError, np.linalg.LinAlgError, TypeError) as e:
            warnings.warn(f"Numerically solving due to the following error: {e}")
            xres, _ = scop.nnls(mprime, b, maxiter=kwargs.get("nnls_max_iter"))

        if removed_index is not None:
            xres = np.insert(xres, removed_index, 1.)

        for index, element in enumerate(self.big_edges_to_use):
            edges_to_use = [list(set(self.frame.vertices[element[vid]].ownEdges) & 
                            set(self.frame.vertices[element[vid+1]].ownEdges))[0]
                            for vid in range(0, len(element)-1)]
            for e in edges_to_use:
                self.frame.edges[e].tension = float(xres[index])

        xres = xres[:-1]
        xres = self.get_solution_no_discarded(xres)
        self.force_dictionary = {}
        # TODO: Reincorporate external forces
        for index, value in enumerate(xres):
        # for i in range(0, tote):
        #     if i < tote:
        #         val = float(xres[i])
            self.force_dictionary[index] = value
        i = 0

        if type(self.externals_to_use) == list or \
            self.externals_to_use == "all" or \
            self.externals_to_use == "ext":
            extForces = {}
            for vid in self.externals_to_use:
                current_border_id = np.where(np.array(self.frame.border_vertices).astype(int) == vid)[0]
                row_idx = self.map_vid_to_row[vid]
                col_idx = int(tote + current_border_id * 2)
                extForces[vid] = self.matrix[row_idx, col_idx: col_idx + 2].tolist()

            for index, _ in extForces.items():
                val1 = round(xres[tote+i], 3) * extForces[index][0]
                val2 = round(xres[tote+i+1], 3) * extForces[index][1]
                self.force_dictionary[f'F{index}x'] = val1
                self.force_dictionary[f'F{index}y'] = val2
                i += 1

        return self.force_dictionary

    def add_mean_one(self, b: np.ndarray) -> Tuple:
        """Add the lagrange multiplier required the average of stresses values equal to one.

        :param b: Right hand side matrix. If in a statical modality it is the null column, \
            in dynamcal modality, has the components of the velocity for each edge.
        :type b: Matrix
        :return: The new LHS and RHS matrices
        :rtype: Tuple
        """
        mprime = self.matrix
        total_edges = mprime.shape[1]
        total_vertices = mprime.shape[0]
        total_borders = len(self.externals_to_use)

        cMatrix = np.array([1.] * total_edges + [0.] * (total_borders * 2))
        mprime = np.vstack((mprime, cMatrix))
        # Now insert the two corresponding cols for the lagrange multipliers
        cMatrix = np.array([1.] * total_vertices + [0.] * (total_borders * 2) + [0.])
        mprime = np.hstack((mprime, cMatrix.reshape(-1, 1)))
        b = np.vstack((b, np.zeros(b.shape[1])))
        b[-1, -1] = total_edges

        if total_borders != 0:
            # cmatrix forces
            cMatrix = np.array([0.] * total_edges + [1.] * (total_borders * 2) + [0.])
            mprime = np.hstack((mprime, cMatrix.reshape(-1, 1)))
            cMatrix = np.concatenate((cMatrix, np.zeros(1)))
            mprime = np.vstack((mprime, cMatrix))

            b = np.vstack((b, np.zeros(b.shape[1])))

        return mprime, b
    
    def add_mean_one_before(self, b: np.ndarray) -> Tuple:
        """Add the lagrange multiplier required the average of stresses values equal to one.

        :param b: Right hand side matrix. If in a statical modality it is the null column, \
            in dynamcal modality, has the components of the velocity for each edge.
        :type b: Matrix
        :return: The new LHS and RHS matrices
        :rtype: Tuple
        """
        mprime = self.matrix.T @ self.matrix
        b = self.matrix.T @ b
        total_edges = mprime.shape[1]
        total_borders = len(self.externals_to_use)

        cMatrix = np.array([1.] * total_edges + [0.] * (total_borders * 2))
        mprime = np.hstack((mprime, cMatrix.reshape(-1, 1)))

        # Now insert the two corresponding cols for the lagrange multpliers
        cMatrix = np.concatenate((cMatrix, np.zeros(1)))
        mprime = np.vstack((mprime, cMatrix))

        b = np.vstack((b, np.zeros(b.shape[1])))
        b[-1, -1] = total_edges

        if total_borders != 0:
            # cmatrix forces
            cMatrix = np.array([0.] * total_edges + [1.] * (total_borders * 2) + [0.])
            mprime = np.hstack((mprime, cMatrix.reshape(-1, 1)))

            cMatrix = np.concatenate((cMatrix, np.zeros(1)))
            mprime = np.vstack((mprime, cMatrix))

            b = np.vstack((b, np.zeros(b.shape[1])))

        return mprime, b

    def fix_one_stress(self, b: np.ndarray, column_to_fix: int = None, value_to_fix_to: float = 1) -> Tuple:
        """Fix one of the stresses values. This is helpful for establishing an scale.

        :param b: Right hand side matrix. Zero in the statical case.
        :type b: Matrix
        :param column_to_fix: Column number to be fixed, i.e edge stress. If None uses \
            the vertex with the most connections, defaults to None
        :type column_to_fix: int, optional
        :param value_to_fix_to: Value to fix the stress to, defaults to 1
        :type value_to_fix_to: float, optional
        :return: New LHS and RHS matrix as well as the index of the fixed stress.
        :rtype: Tuple
        """
        # Replace column with the most connections
        if not column_to_fix:
            non_zero_count = np.count_nonzero(self.matrix, axis=0)
            max_index = np.argmax(non_zero_count)
            b = b - value_to_fix_to * self.matrix[:, max_index]
            self.matrix = np.delete(self.matrix, max_index, 1)
        else:
            raise(NotImplementedError)
        
        mprime = self.matrix.T @ self.matrix
        b = self.matrix.T @ b

        return mprime, b, max_index

    def set_velocity_matrix(self, timeseries: Union[str, list] = None, **kwargs):
        vector_of_vectors = []
        b = np.zeros((self.matrix.shape[0], 1))
        b_matrix = kwargs.get("b_matrix", None)
        if timeseries and (b_matrix == "velocity" or b_matrix == "acceleration"):
            for vid in self.map_vid_to_row.keys():
                if b_matrix == "velocity":
                    value = timeseries.calculate_velocity(vid, self.frame.frame_id)
                elif b_matrix == "acceleration":
                    value = timeseries.calculate_acceleration(vid, self.frame.frame_id)
                    if np.any(np.isnan(value)):
                        value = [0, 0]
                else:
                    raise NotImplementedError(f'{b_matrix=}')
                j = self.map_vid_to_row[vid]
                b[j, 0] = value[0]
                b[j + 1, 0] = value[1]
                vector_of_vectors.append(value)

        self.velocity_normalization = kwargs.get("velocity_normalization", 1)
        if len(vector_of_vectors) != 0 and kwargs.get("adimensional_velocity",  False):
            average_velocity = np.mean([np.linalg.norm(vector) for vector in vector_of_vectors])
        else:
            average_velocity = 1

        self.velocity_matrix_dimensional = b.T.astype(np.float64).round(4)
        
        b = (b / average_velocity) * self.velocity_normalization
        
        self.velocity_matrix = b.T.astype(np.float64).round(4)

        return b, average_velocity
    
    def get_solution_no_discarded(self, xres: np.ndarray) -> list:
        """Get new solution adding -1 to the big edges that no longer exist

        :param x0: Original solution
        :type x0: list
        :return: New list with solution for the inexesitant edges equal to -1
        :rtype: list
        """
        big_edges_vertices = [big_edge.get_vertices_ids() for big_edge in self.frame.internal_big_edges]
        if len(big_edges_vertices) == len(xres):
            return xres

        xres_i = 0  # pointer to next value in xres
        xres_new = np.empty(len(big_edges_vertices), dtype=xres.dtype)
        for be_index, big_edge in enumerate(big_edges_vertices):
            if (big_edge[0] in self.deletes) and (big_edge[-1] in self.deletes):
                xres_new[be_index] = -1
            else:
                xres_new[be_index] = xres[xres_i]
                xres_i += 1
        return xres_new
    
    def get_new_initial_condition(self, x0: list, what="zero") -> list:
        """Get new initial condition adding zeros to or removing the big edges that no longer exist

        :param x0: Original list
        :type x0: list
        :return: New list with initial condition for the inexesitant edges equal to zero
        :rtype: list
        """
        list_of_big_edges = [big_edge.get_vertices_ids() for big_edge in self.frame.internal_big_edges]
        both_count = 0
        removed_indices = {}
        for index, big_edge in enumerate(list_of_big_edges):
            if (big_edge[0] in self.deletes) and (big_edge[-1] in self.deletes):
                removed_indices[index] = x0[index]
                if what == "zero":
                    x0[index] = 0
                else:
                    x0[index] = -1
                both_count += 1
        return x0, removed_indices


from dataclasses import dataclass
import pandas as pd
import os
import re
from typing import Tuple

import forsys.vertex as vertex
import forsys.edge as edge
import forsys.cell as cell

@dataclass
class SurfaceEvolver:
    """Class interface with SurfaceEvolver-generated files
    
    :param fname: Path to the SurfaceEvolver file to read
    :type fname: str
    """
    fname: str


    def __post_init__(self):
        self.vertices, self.edges, self.cells = self.create_lattice()

    def create_lattice(self) -> Tuple:
        """
        Create vertices, edges and cells from a Surface Evolver file. 
        All necessary steps are taken by this call.

        :return: Three dictionaries with the vertices, edges and cells respectively   
        :rtype: Tuple
        """
        edges_temp = self.get_edges()
        
        vertices = {}
        for _, r in self.get_vertices().iterrows():
            vertices[int(r.id)] = vertex.Vertex(int(r.id), r.x, r.y)
            
        edges  = {}
        for _, r in edges_temp.iterrows():
            edges[int(r.id)] = edge.SmallEdge(int(r.id), vertices[int(r.id1)], vertices[int(r.id2)])
            edges[int(r.id)].gt = round(edges_temp.loc[edges_temp['id'] == int(r.id)]['force'].iloc[0], 4)

        cells = {}
        for _, r in self.get_cells().iterrows():
            vlist = [edges[abs(e)].v1 if e > 0 else edges[abs(e)].v2 for e in r.edges]
            gt_pressure = round(r["pressures"], 4)
            cells[int(r.id)] = cell.Cell(int(r.id), vlist, gt_pressure=gt_pressure)

        # delete all vertices and edges with no cell
        vertex_to_delete = []
        for vid, v in vertices.items():
            if len(v.ownCells) == 0:
                vertex_to_delete.append(int(vid))

        for i in vertex_to_delete:
            for e in vertices[i].ownEdges.copy():
                try:
                    del edges[e] 
                except KeyError:
                    pass
        
            del vertices[i]

        return vertices, edges, cells

    def calculate_first_last(self) -> Tuple:
        """
        Calculate starting position of the vertices, edges, cells and pressures
        definitions in the Surface Evolver file.

        :return: Four tuples with the initial and final line index for the \
            vertices, edges, cells and pressures respectively.
        :rtype: Tuple
        """
        with open(os.path.join(self.fname),"r") as f:
            ini_v = next(index for index, line in enumerate(f) if line.startswith("vertices  "))
            fin_v = next(index for index, line in enumerate(f) if line.startswith("edges  ")) + ini_v
            f.seek(0)
            ini_e = next(index for index, line in enumerate(f) if line.startswith("edges  "))
            fin_e = next(index for index, line in enumerate(f) if line.startswith("faces  ")) + ini_e
            f.seek(0)
            ini_f = next(index for index, line in enumerate(f) if line.startswith("faces  "))
            fin_f = next(index for index, line in enumerate(f) if line.startswith("bodies  ")) + ini_f
            f.seek(0)
            ini_p = next(index for index, line in enumerate(f) if line.startswith("bodies  "))
            fin_p = next(index for index, line in enumerate(f) if line.startswith("read")) + ini_p

        self.index_v = (ini_v, fin_v)
        self.index_e = (ini_e, fin_e)
        self.index_f = (ini_f, fin_f)
        self.index_pressures = (ini_p, fin_p)

        return self.index_v, self.index_e, self.index_f, self.index_pressures


    def get_first_last(self) -> Tuple:
        """
        Get list of line indices for all conditions. If the lines are already known,
        returns known values. If not, sets them and the returns their values.

        :return: Four tuples with the initial and final line index for the \
            vertices, edges, cells and pressures respectively.
        :rtype: Tuple
        """
        try:
            return self.index_v, self.index_e, self.index_f, self.index_pressures
        except AttributeError:
            return self.calculate_first_last()
    
    def get_vertices(self) -> pd.DataFrame:
        """
        Generate DataFrame with the system's vertices IDs and positions

        :return: Dataframe with three columns: ID, x and y for each vertex in \
            the system.
        :rtype: pd.DataFrame
        """
        vertices = pd.DataFrame()
        ids = []
        xs = []
        ys = []

        index_v, _, _, _ = self.get_first_last()

        with open(os.path.join(self.fname), "r") as f:
            lines = f.readlines()
            for i in range(index_v[0] + 1, index_v[1]):
                ids.append(int(re.search(r"\d+", lines[i]).group()))
                xs.append(round(float(lines[i].split()[1]), 3))
                ys.append(round(float(lines[i].split()[2]), 3))
                # vals.append(float(lines[i][lines[i].find("density"):].split(" ")[1]))
        vertices['id'] = ids
        vertices['x'] = xs
        vertices['y'] = ys

        return vertices
    
    def get_edges(self) -> pd.DataFrame:
        """
        Generate DataFrame with the system's edges IDs, the two vertices that 
        form them and the ground truth density.

        :return: Dataframe with four columns: ID, vertex 1, vertex 2 \
            and ground truth density for each edge.
        :rtype: pd.DataFrame
        """
        edges = pd.DataFrame()
        ids = []
        id1 = []
        id2 = []
        forces = []

        _, index_e, _, _ = self.get_first_last()

        with open(os.path.join(self.fname), "r") as f:
            lines = f.readlines()
            for i in range(index_e[0] + 1, index_e[1]):
                ids.append(int(re.search(r"\d+", lines[i]).group()))
                id1.append(int(lines[i].split()[1]))
                id2.append(int(lines[i].split()[2]))
                forces.append(float(lines[i].split()[4]) if lines[i].split()[3] == "density" else 1)
                # vals.append(float(lines[i][lines[i].find("density"):].split(" ")[1]))
        edges['id'] = ids
        edges['id1'] = id1
        edges['id2'] = id2
        edges['force'] = forces
        return edges

    def get_cells(self) -> pd.DataFrame:
        """
        Generate DataFrame with the system's cells IDs, its edges, 
        and their ground truth pressures.

        :return: Dataframe with three columns: ID, array of edges from the cell \
            and the ground truth pressure.
        :rtype: pd.DataFrame
        """
        cells = pd.DataFrame()
        ids = []
        edges = []

        pressure_dict = self.get_pressures()

        _, _, index_f, _ = self.get_first_last()

        with open(os.path.join(self.fname), "r") as f:
            lines = f.readlines()
            current_edge = []
            first = True

            for i in range(index_f[0] + 1, index_f[1]):
                splitted = lines[i].split()
                if first and "*/" not in splitted[-1]:
                    ids.append(splitted[0])
                    first = False
                    current_edge = current_edge+splitted[1:-1]
                elif splitted[-1] == '\\' and not first:
                    current_edge = current_edge+splitted[0:-1]
                elif "*/" in splitted[-1]:
                    # last line, save
                    if first:
                        ids.append(splitted[0])
                        current_edge = current_edge+splitted[1:-2]
                    else:
                        current_edge = current_edge+splitted[0:-2]
                    edges.append([int(e) for e in current_edge])
                    current_edge = []
                    first = True
        
        cells['id'] = ids
        cells['edges'] = edges
        cells["pressures"] = pressure_dict.values()

        return cells
    
    def get_pressures(self) -> dict:
        """Get the values of the ground truth pressure of the cells

        :return: Ground truth pressure dictionary.
        :rtype: dict
        """
        pressures = {}
        _, _, _, index_p = self.get_first_last()
        with open(os.path.join(self.fname), "r") as f:
            lines = f.readlines()

            for i in range(index_p[0] + 1, index_p[1]):
                splitted = lines[i].split()
                pressures[int(splitted[0])] = float(splitted[7])
        return pressures




import itertools as itertools
import numpy as np
from collections import Counter
from typing import Tuple

import scipy.optimize as sco
import circle_fit as cfit
import forsys.vertex as fvertex
import forsys.edge as fedge

from forsys.exceptions import SegmentationArtifactException, BigEdgesBadlyCreated


def create_edges_new(vertices, cells):
    """
    Create big edges by locating junctions with more that 2 lines connecting
    """
    earr_temp = []
    for cell in cells.values():
        # vertices = {v.id: v for v in cell.vertices}
        cell_vertex_ids = [v.id for v in cell.vertices]
        number_of_connections = [len(n.ownEdges) > 2 for n in cell.vertices]

        partitioned = get_partition(cell_vertex_ids, number_of_connections)
        if not number_of_connections[0]:
            new_vertex_order = np.concatenate((*partitioned[1:], partitioned[0]))
            new_number_of_connections = [len(vertices[vid].ownEdges) > 2 for vid in new_vertex_order]
            partitioned = get_partition(new_vertex_order, new_number_of_connections)

        partitioned = partitioned[1:]
        new_edges = [np.append(partitioned[ii], partitioned[((ii + 1) % len(partitioned))][0]).flatten().tolist() for ii in
                     range(0, len(partitioned))]

        earr_temp += new_edges

    earr = []
    for e in earr_temp:
        if e[::-1] not in earr and e not in earr:
            earr += [e]
    return earr


def get_partition(cell_vertex_ids, number_connections):
    indices = [i for i in np.where(number_connections)[0]]
    partitioned = np.split(cell_vertex_ids, indices)
    return partitioned


def generate_mesh(vertices, edges, cells, ne=4, **kwargs):
    # bedges = create_edges(cells)
    bedges = create_edges_new(vertices, cells)
    # ldict = {}
    nEdgeArray = []
    alreadySeen = []
    vertices_to_join = []
    edgeRange = range(0, ne)
    # print(bedges)
    for e in bedges:
        if len(e) > ne:
            if not e in alreadySeen or not reversed(e) in alreadySeen:
                each = len(e) / ne
                nEdge = []
                for i in edgeRange:
                    nEdge.append(e[int(each * i)])
                nEdge.append(e[-1])
                nEdgeArray.append(nEdge)
                alreadySeen.append(e)
        else:
            nEdgeArray.append(e)
            alreadySeen.append(e)
            # if edge has only 2 vertices, join them 
            # TODO: Make it work with polygonal vertex model
            if (len(e) == 2 and
                len(vertices[e[0]].ownCells) < 3 and
                len(vertices[e[1]].ownCells) < 3):
                vertices_to_join.append(e)

    vertexToRemove = []
    for vID, v in vertices.items():
        if not vID in list(itertools.chain.from_iterable(nEdgeArray)):
            if not vID in vertexToRemove:
                vertexToRemove.append(vID)
            for cid in v.ownCells:
                cells[cid].vertices.remove(v)
                # cellsToRemove.append(cid)

    # remove all edges
    edges.clear()
    for vi in vertexToRemove:
        del vertices[vi]
    #### from here
    edges = {}
    # now join the vertices
    edgesNumber = 0
    for be in nEdgeArray:
        rango = range(0, len(be) - 1)
        for n in rango:
            edges[edgesNumber] = fedge.SmallEdge(edgesNumber, vertices[be[n]], vertices[be[n + 1]])
            edgesNumber += 1

    # if there is an empty cell, may be its an artifact, remove it
    cells_to_remove = []
    for c in cells.keys():
        if len(cells[c].vertices) == 0:
            cells_to_remove.append(c)

    for c in cells_to_remove:
        print(f"*** WARNING **** CELL {cells[c].id} remove due to having {len(cells[c].vertices)} vertices")
        del cells[c]

    # if len(vertices_to_join) > 0 and kwargs.get("replace_short_edges", True):
    mapper = {}
    if kwargs.get("replace_short_edges", True):
        for edge_to_join in vertices_to_join:
            try:
                vertices, edges, cells, new_mapper = join_two_vertices(edge_to_join, vertices, edges, cells, mapper)
                mapper = mapper | new_mapper
            except KeyError:
                raise SegmentationArtifactException()
    # exit()

    return vertices, edges, cells, nEdgeArray


def eid_from_vertex(earr, vbel):
    # ret = []
    for j in range(0, len(earr)):
        if len(list(set(earr[j]) & set(vbel))) >= 2:
            return j
    raise BigEdgesBadlyCreated()


def get_border_edge(earr, vertices):
    borderEdge = []
    for eid in earr:
        # print("Edge ", eid)
        for vid in eid:
            if len(vertices[vid].ownCells) < 2:
                borderEdge.append(eid)
                break
    return borderEdge


def get_border_from_angles(earr, vertices):
    inBorder = []
    bedge = get_border_edge(earr, vertices)
    for eid in bedge:
        # Choose pivot vector
        anglesArr = []
        for i in range(1, len(eid) - 1):
            v1 = (vertices[eid[i - 1]].x - vertices[eid[i]].x,
                  vertices[eid[i - 1]].y - vertices[eid[i]].y)
            v2 = (vertices[eid[i + 1]].x - vertices[eid[i]].x,
                  vertices[eid[i + 1]].y - vertices[eid[i]].y)

            anglesArr.append(angle_between_two_vectors(v1, v2))
        diff = 0
        for a, b in itertools.combinations(anglesArr, 2):
            if abs(a - b) > 0.3:
                diff += 1
        if diff == 2 or diff == 0:
            inBorder.append(eid[int(len(eid) / 2)])
        else:
            inBorder.append(eid[1 + anglesArr.index(min(anglesArr))])
    return inBorder


def get_border_from_angles_new(earr, vertices):
    # put the one in the middle
    inBorder = []
    for eid in get_border_edge(earr, vertices):
        # for i in range(1, len(eid) - 1):
        inBorder.append(eid[int(len(eid) / 2)])
    return inBorder


def get_big_edges_se(frame):
    """
    Get the big edges by detecting changes in ground truth of Surface Evolver
    """
    new_big_edges = []
    middle_vertices = []

    for bedge in frame.earr:
        edges_to_use = frame.get_big_edge_edgesid(bedge)
        edges_tension = [frame.edges[eid].gt for eid in edges_to_use]

        a = [1 if edges_tension[b] != edges_tension[b + 1] else 0 for b in range(0, len(edges_tension) - 1)]

        cut_where = np.where(np.array(a).astype(int) == 1)[0] + 1
        new_edges = np.split(bedge, cut_where)
        for enumber in range(0, len(new_edges) - 1):
            new_big_edges.append(list(np.concatenate((new_edges[enumber], [new_edges[enumber + 1][0]]))))
            middle_vertices.append(new_edges[enumber + 1][0])
        new_big_edges.append(list(new_edges[-1]))

    return new_big_edges, middle_vertices


def non_border_big_edges(inBorder, earr):
    newEarr = []
    for eid in earr:
        border = False
        for vid in eid:
            if vid in inBorder:
                newEarr.append(earr[earr.index(eid)][:eid.index(vid)] + [vid])
                newEarr.append(earr[earr.index(eid)][eid.index(vid):])
                border = True
        if not border:
            newEarr.append(eid)
    return newEarr


def get_virtual_edges(earrN, vertices):
    vseen = []
    virtualEdges = {}
    edges = []
    for eid in earrN:
        if eid[0] not in vseen:
            vseen.append(eid[0])
    for vid in vseen:
        virtualEdges[vid] = vertices[vid].ownEdges
        edges.append(virtualEdges[vid])
    edges = set(list(itertools.chain.from_iterable(edges)))

    return virtualEdges, edges


def get_edge_from_earr(edges, element):
    for eid, eobj in edges.items():
        if Counter(eobj.get_vertices_id()) == Counter(element):
            return eid


def get_line_constant_from_bedge(edges, ele):
    edgeLC = []
    # print("Element: ", ele)
    for i in range(0, len(ele) - 1):
        # print("Slice: ", ele[i:i+2])
        eid = get_edge_from_earr(edges, ele[i:i + 2])
        edgeLC.append(edges[eid].getLineConstant())
    return edgeLC


def calculate_circle_center(vertices, method: str="dlite") -> Tuple:
        """
        Calculate the centroid of the cell through a circle fit 
        dlite invokes a modified from DLITE's implementation: 
        https://github.com/AllenCellModeling/DLITE
        if no argument is passed, the geometric center is returned

        :return: x and y coordiante of the cell's center from a circular fit
        :rtype: Tuple
        """        
        # print("Using as method: ", method)
        xs = [v.x for v in vertices]
        ys = [v.y for v in vertices]

        if method == "dlite":
            center = dlite_circle_method(xs, ys)
        elif method == "taubinSVD":
            zipped_coords = list(zip(xs, ys))
            if len(zipped_coords) < 3:
                center = dlite_circle_method(xs, ys)
            else:
                try:
                    xc, yc, r, sigma = cfit.taubinSVD(zipped_coords)
                    center = [xc, yc]
                except FloatingPointError:
                    center = dlite_circle_method(xs, ys)
        else:
            center = [np.mean(xs),  np.mean(ys)]
        
        return center[0], center[1]

def dlite_circle_method(xs, ys):
    def objective_f(c):
        """ 
        Distance between the vertices and the mean circle centered at c
        """
        distances = np.sqrt((xs - c[0]) ** 2 + (ys - c[1]) ** 2)
        return distances - distances.mean()

    center, _ = sco.leastsq(objective_f, (np.mean(xs), np.mean(ys)))
    return center



def get_versors(vertices, edges, vid):
    for eb in vertices[vid].ownEdges:
        # get versors average of the next two
        edge = edges[eb]
        deltaX = edge.v2.x - edge.v1.x
        deltaY = edge.v2.y - edge.v1.y
        modulus = np.sqrt(deltaX ** 2 + deltaY ** 2)
        # check if the edge is in the border, and don't create a lambda if it is
        other_vertex = list(set(edge.get_vertices_id()) - set([vid]))
        if len(other_vertex) == 1:
            border = len(vertices[other_vertex[0]].ownCells) == 1
        else:
            border = False
        if vid == edge.get_vertices_id()[0]:
            vx = deltaX / modulus
            vy = deltaY / modulus
        else:
            vx = - deltaX / modulus
            vy = - deltaY / modulus
        # Force versor, to calculate modulus
        yield (vx, vy, edge.get_vertices_id(), border, edge.id)


def get_border_edges(earr, vertices, edges):
    border_edges = []
    for e in earr:
        if len(vertices[e[int(len(e) / 2)]].ownCells) == 1:
            border_edges.append(e)
    return border_edges


def get_versors_average(vertices, edges, vid):
    for eb in vertices[vid].ownEdges:
        edge = edges[eb]

        other_vertex_id = list(set(edge.get_vertices_id()) - set([vid]))

        if len(other_vertex_id) == 1:
            border = len(vertices[other_vertex_id[0]].ownCells) == 1
        else:
            border = False

        # Next vertex
        next_edge = edges[list(set(vertices[other_vertex_id[0]].ownEdges) - set([eb]))[0]]

        vector1 = edge.get_vector()

        sign1 = 1
        sign2 = 1
        if vid != edge.get_vertices_id()[0]:
            sign1 = -1
        if next_edge.get_vertices_id()[0] != other_vertex_id[0]:
            sign2 = -1

        vector2 = next_edge.get_vector()

        ret_versor = sign1 * np.array(vector1) + sign2 * np.array(vector2)
        ret_versor = ret_versor / np.linalg.norm(ret_versor)

        # yield (ave_versor[0], ave_versor[1], edge.get_vertices_id(), border)
        yield (ret_versor[0], ret_versor[1], edge.get_vertices_id(), border)


def angle_between_two_vectors(a, b):
    clipped_dot_product = np.clip(np.dot(a / np.linalg.norm(a), b / np.linalg.norm(b)), -1, 1)
    return np.arccos(clipped_dot_product)


# def replace_short_edges(vertices_to_join, vertices, edges, cells):
def join_two_vertices(vertices_to_join, vertices, edges, cells, mapper={}):
    """ Join the two-vertex edges"""
    try:
        v0 = vertices[vertices_to_join[0]]
    except KeyError:
        v0 = vertices[mapper[vertices_to_join[0]]]
    try:
        v1 = vertices[vertices_to_join[1]]
    except KeyError:
        v1 = vertices[mapper[vertices_to_join[1]]]
    
    x_cm = abs(v0.x + v1.x) / 2
    y_cm = abs(v0.y + v1.y) / 2

    # find the common edge
    common_edge = list(set(v0.ownEdges) & set(v1.ownEdges))[0]

    new_id = get_unused_id(vertices)        
    new_vertex = fvertex.Vertex(new_id, x_cm, y_cm)
    vertices[new_id] = new_vertex
    mapper[vertices_to_join[0]] = new_id
    mapper[vertices_to_join[1]] = new_id
    # replace vertex in the cells
    list_of_cells_0 = [cid for cid in v0.ownCells]
    list_of_cells_1 = [cid for cid in v1.ownCells]
    for cid in list_of_cells_0:
        cells[cid].replace_vertex(vertices[v0.id], new_vertex)
    for cid in list_of_cells_1:
        cells[cid].replace_vertex(vertices[v1.id], new_vertex)
    # destroy edge
    del edges[common_edge]
    # replace vertex in the edges
    list_of_edges_0 = [eid for eid in v0.ownEdges]
    list_of_edges_1 = [eid for eid in v1.ownEdges]
    for edge_id in list_of_edges_0:
        edges[edge_id].replace_vertex(vertices[v0.id], new_vertex)
    for edge_id in list_of_edges_1:
        edges[edge_id].replace_vertex(vertices[v1.id], new_vertex)

    # destroy the two previous vertices
    del vertices[v0.id]
    del vertices[v1.id]

    return vertices, edges, cells, mapper


def get_unused_id(dictionary):
    new_id = len(dictionary)
    i = 0
    while dictionary.get(new_id) != None:
        new_id = len(dictionary) + i
        i += 1
    return new_id


import forsys.general_matrix as forsys_general_matrix
import numpy as np
from typing import Tuple


class PressureMatrix(forsys_general_matrix.GeneralMatrix):
    """Child class to manage creating of the matrix system of equations
    to calculate pressure. The corresponding stresses need to be solved already.
    """
    
    def __init__(self, frame: object, timeseries: dict):
        """Constructor method

        :param frame: Frame object from which matrix system is created
        :type frame: object
        :param timeseries: Time series dictionary of frames.
        :type timeseries: dict
        """
        super().__init__(frame, timeseries)
        self._build_matrix()

    def _build_matrix(self) -> Tuple:
        """Generate the matrix system for pressure inference. 

        :return: Returns LHS matrix and RHS matrix
        :rtype: Tuple
        """
        n_edges, n_cells = len(self.big_edges_to_use), len(self.frame.cells)
        self.mapping_order = {key: enumid for enumid, key in enumerate(self.frame.cells)}

        self.rhs_matrix = np.empty(n_edges)
        self.lhs_matrix = np.empty((n_edges, n_cells))
        for position_id, big_edge in enumerate(self.big_edges_to_use):
            lhs_row, rhs_value = self.get_row(big_edge)
            self.rhs_matrix[position_id] = rhs_value
            self.lhs_matrix[position_id] = lhs_row

        # remove columns with all zeros
        self.removed_columns = np.nonzero(np.all(self.lhs_matrix == 0, axis=0))[0].tolist()  # list of column indices
        if self.removed_columns:
            self.lhs_matrix = np.delete(self.lhs_matrix, self.removed_columns, axis=1)

        return self.lhs_matrix, self.rhs_matrix


    def get_row(self, big_edge: object) -> Tuple:
        """Creates the row corresponding to a given big_edge for the LHS matrix 
        and the RHS one. The stress needs to be already inferred.

        :param big_edge: Object instance of the big edge that this row would correspond
        :type big_edge: object
        :return: Returns the left hand side row and the right hand side value.
        :rtype: Tuple
        """
        lhs_row = np.zeros(len(self.mapping_order))  # n_cells
        big_edge_cells = big_edge.own_cells
        if len(big_edge_cells) != 2:
            raise ValueError(f'big edge has own_cells={len(big_edge_cells)}, expecting 2')

        curvature = big_edge.calculate_total_curvature(normalized=False)
        rhs_value = big_edge.tension * curvature

        c1_position = self.mapping_order[big_edge_cells[0]]
        c2_position = self.mapping_order[big_edge_cells[1]]

        if self.frame.cells[big_edge_cells[0]].get_area_sign() > 0:
            lhs_row[c1_position] = 1
            lhs_row[c2_position] = -1
        else:
            lhs_row[c1_position] = -1
            lhs_row[c2_position] = 1

        return lhs_row, rhs_value



from dataclasses import dataclass
import numpy as np
from PIL import Image
import cv2
import itertools as iter
from collections import Counter
from typing import Tuple

import forsys.vertex as vertex
import forsys.edge as edge
import forsys.cell as cell
import forsys.wkt as fwkt
import forsys.virtual_edges as fvedges

@dataclass
class Skeleton:
    """
    Reads and parses TIFF skekeletons using OpenCV implementation to detect polygons

    :param fname: Path to segmented image
    :type fname: str
    :param mirror_y: False by default. Inverts the y-axis if true.
    :type mirror_y: bool, optional
    """
    fname: str

    mirror_y: bool = False

    def __post_init__(self):
        with Image.open(self.fname).convert("L") as curr_image_file:
            self.img_arr = np.array(curr_image_file)[1:-1, 1:-1]

        # self.contours, _ = cv2.findContours(self.img_arr, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        self.contours, _ = cv2.findContours(self.img_arr, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
        self.contours = [np.vstack(p).squeeze() for p in self.contours][1:]
        # check size of all areas and filter small ones
        all_cell_areas = [self.calculate_area(polygon) for polygon in self.contours]
        # if a cell has more than 5 times the average cell area (withouth the maximum)
        # remove it
        ave_cell_area = np.mean(np.sort(all_cell_areas)[:-1])
        self.contours = [polygon for polygon in self.contours if self.calculate_area(polygon) < 5 * ave_cell_area]
        # initialize the counters
        self.vertex_id = 0
        self.edge_id = 0
        self.cell_id = 0

        if self.mirror_y:
            self.max_y = float(max(list(iter.chain.from_iterable(
                [list(zip(*c))[1] for c in self.contours]))))

    def create_lattice(self, **kwargs) -> Tuple[dict, dict, dict]:
        """
        Creates vertices, edges and cells of the system in three dictioinaries

        :return: Vertices, edges and cells created from parsing the provided image
        :rtype: tuple
        """
        self.vertices = {}
        self.edges = {}
        self.cells = {}

        self.coords_to_key = {}

        self.edges_added = []

        rescale = kwargs.get("rescale", [1, 1])
        offset = kwargs.get("offset", [0, 0])

        for _, polygon in enumerate(self.contours):
            cell_vertices_list = []
            for coords in polygon:
                if self.mirror_y:
                    coords[1] = self.max_y - coords[1]
                vid = self.get_vertex_id_by_position(coords)

                if vid == -1:
                    vid = self.vertex_id
                    self.vertices[vid] = vertex.Vertex(int(vid),
                                                       (coords[0] - offset[0]) / rescale[0],
                                                       (coords[1] - offset[1]) / rescale[1])
                    self.vertex_id += 1

                    self.coords_to_key[tuple(coords)] = vid
                    # add vertex to this cell's list
                cell_vertices_list.append(self.vertices[vid])

            # join vertices with the edges
            number_of_vertices = len(polygon)
            for n in range(1, number_of_vertices):
                self.create_edge(n - 1, n, polygon)
            # Join last vertex with first
            self.create_edge(n, 0, polygon)

            # create cells now
            self.cells[self.cell_id] = cell.Cell(self.cell_id, cell_vertices_list, {})
            self.cell_id += 1
        
        if kwargs.get("reduce_amount", False):
            self.vertices, self.edges, self.cells = fwkt.reduce_amount(self.vertices,
                                                                    self.edges,
                                                                    self.cells)

        self.all_big_edges = fvedges.create_edges_new(self.vertices,
                                                      self.cells)

        for current_cell in self.cells.values():
            # if any of the vertices in the cell has a vertex with only one cell, that is an external cell
            if np.any([len(v.ownCells) == 1 for v in current_cell.vertices]):
                current_cell.is_border = True
        for e in self.edges.values():
            # if at least one of the vertices have only one cell, this is an external edge
            if len(e.v1.ownCells) == 1 or len(e.v2.ownCells) == 1:
                e.external = True
            else:
                e.external = False

        # triangles in the middle
        inner_edge_triangles = []
        num_elements = len(self.all_big_edges)
        first_part = [(e[0], e[-1]) for e in self.all_big_edges]
        last_part = [(e[-1], e[0]) for e in self.all_big_edges]
        first_last = first_part + last_part
        inner_edge_triangles = [k for k, v in Counter(first_last).items() if v > 1]
        
        visited = []
        for index in range(len(inner_edge_triangles) - 1):
            if (inner_edge_triangles[index] in visited or
                inner_edge_triangles[index][::-1] in visited):
                continue

            edge_0_index = first_last.index(inner_edge_triangles[index]) % num_elements
            edge_1_index = first_last.index(inner_edge_triangles[(index + 1)]) % num_elements
            edge_0 = self.all_big_edges[edge_0_index]
            if len(edge_0) > 3:
                continue
            edge_1 = self.all_big_edges[edge_1_index]
            vertex_id_to_delete = np.setdiff1d(edge_0, edge_1)[0]

            its_cells = self.vertices[vertex_id_to_delete].ownCells
            for cell_id in its_cells:
                self.cells[cell_id].replace_vertex(self.vertices[vertex_id_to_delete],
                                                   self.vertices[edge_0[0]])

            its_edges = self.vertices[vertex_id_to_delete].ownEdges
            for edge_id in its_edges:
                del self.edges[edge_id]

            del self.vertices[vertex_id_to_delete]
            visited.append(inner_edge_triangles[index])

        # get artifacts from the contour and apply T3 transitions to them
        all_artifact_vertices = list(self.get_artifacts())

        artifacts = []
        while len(all_artifact_vertices) != 0:
            current_artifact = [all_artifact_vertices[0]]
            current_len0 = -1
            while current_len0 != len(current_artifact):
                current_artifact = self.add_vertices_to_current(all_artifact_vertices, current_artifact)
                current_len0 = len(current_artifact)
            # remove vertices already in artifact from all_artifact_vertices
            for ii in current_artifact:
                all_artifact_vertices.remove(ii)

            artifacts.append(current_artifact)

        # now make a T3 transition for each triangle:
        for artifact in artifacts:
            self.do_t3_transition(artifact)

        # if all vertices in a cell only have that cell as own, delete cell
        cells_to_remove = []
        for c in self.cells.values():
            vertices_le_one = [len(v.ownCells) <= 1 for v in c.vertices]

            if np.all(vertices_le_one):
                cells_to_remove.append(c.id)
                for v in c.vertices:
                    # remove edges
                    for e in v.ownEdges:
                        del self.edges[e]
                    # remove vertices
                    try:
                        del self.vertices[v.id]
                    except KeyError:
                        print(f"**WARNING** vertex {v.id} not in list of vertices")
        # remove the cell
        for cid in cells_to_remove:
            del self.cells[cid]

        return self.vertices, self.edges, self.cells

    def add_vertices_to_current(self, artifact_vertices: list, current_artifact: list) -> list:
        """
        Creates the list of vertices of a given artifact
        :param artifact_vertices: All vertices belonging to artifacts
        :type artifact_vertices: list
        :param current_artifact: List to append the vertices of the current artifact
        :type current_artifact: list

        :return: Vertices, edges and cells created from parsing the provided image
        :rtype: list
        """
        v0 = current_artifact[-1]
        for eid in self.vertices[v0].ownEdges:
            other_vertex = self.edges[eid].get_other_vertex_id(v0)
            if other_vertex in artifact_vertices and other_vertex not in current_artifact:
                current_artifact.append(other_vertex)
        return current_artifact

    def do_t3_transition(self, artifact: list):
        """
        Perform a T3 transition in all the identified artifacts

        :param artifact: List with all the artifacts detected in the system
        :type artifact: list
        :return: None
        """
        # get the mid point and create a vertex there
        ave_x = np.mean([self.vertices[ii].x for ii in artifact])
        ave_y = np.mean([self.vertices[ii].y for ii in artifact])

        new_vid = self.get_new_vid()
        self.vertices[new_vid] = vertex.Vertex(int(new_vid),
                                               ave_x,
                                               ave_y)

        # get the edge and join them to the new point)
        for v in artifact:
            edge_to_remove = []
            edge_to_replace = []
            for e in self.vertices[v].ownEdges:
                if (self.edges[e].v1.id in artifact and self.edges[e].v2.id in artifact) or \
                        (self.edges[e].v2.id in artifact and self.edges[e].v1.id in artifact):
                    edge_to_remove.append(e)
                else:
                    edge_to_replace.append(e)

            for ii in range(0, len(edge_to_remove)):
                del self.edges[edge_to_remove[ii]]

            for ii in range(0, len(edge_to_replace)):
                edge_object = self.edges[edge_to_replace[ii]]
                # if new_vid != edge_object.v1.id and new_vid != edge_object.v2.id:
                edge_object.replace_vertex(self.vertices[v], self.vertices[new_vid])

            # add the cells corresponding
            for c in self.vertices[v].ownCells:
                # cell_vertex_ids = [v.id for v in self.cells[c]]
                self.cells[c].replace_vertex(self.vertices[v], self.vertices[new_vid])

        for jj in range(0, len(artifact)):
            if len(self.vertices[artifact[jj]].ownEdges) == 0:
                del self.vertices[artifact[jj]]
            else:
                print("WARNING: BAD VERTEX DELETION IN TRIANGLE - MORE THAN ZERO EDGE")

    def get_artifacts(self) -> list:
        """
        Create list with all the artifacts detected in the processing of the tissue

        :return: List of vertices belonging to artifacts
        :rtype: list
        """
        # triangles with TJs
        candidate_artifact_vertices = []
        for v in self.vertices.values():
            if len(v.ownEdges) == 3 and len(v.ownCells) == 2:
                # this is part of a triangle
                candidate_artifact_vertices.append(v.id)

        # list of external vertices:
        external_vertices = []
        for e in self.edges.values():
            if e.external:
                external_vertices.append(e.v1.id)
                external_vertices.append(e.v2.id)
        artifact_vertices = np.setdiff1d(candidate_artifact_vertices, external_vertices)

        return artifact_vertices

    def create_edge(self, n0: int , n1: int, polygon: list):
        """
        Add tuples of vertices identified by the vertices that from them and create the SmallEdge object

        :param n0: ID in the polygon list of the first vertex
        :type n0: int
        :param n1: ID in the polygon list of the second vertex
        :type n1: int
        :param n0: List of vertices of the current polygon
        :type polygon: int
        """
        # TODO: Functionalize function.
        v1 = self.vertices[self.get_vertex_id_by_position(polygon[n0])]
        v2 = self.vertices[self.get_vertex_id_by_position(polygon[n1])]
        if not (v1.id, v2.id) in self.edges_added and not (v2.id, v1.id) in self.edges_added:
            self.edges[self.edge_id] = edge.SmallEdge(self.edge_id, v1, v2)
            self.edge_id += 1
            self.edges_added.append((v1.id, v2.id))

    def get_vertex_id_by_position(self, coordinates: list) -> int:
        """
        Check whether vertices were already created using a hashmap
        
        :param coordinates: tuple with the coordiantes of the vertex to check
        :type coordinates: list

        :return: Current key of the given coordinates or -1 if there are no vertices at that position
        :rype: int
        """
        try:
            return self.coords_to_key[tuple(coordinates)]
        except KeyError:
            return -1

    def get_new_vid(self) -> int:
        """
        Returns the best possible ID for the creation of a new vertex

        :return: unused ID in the vertices dictionary
        :rtype: int
        """
        new_vid = max(self.vertices.keys()) + 1
        while new_vid in self.vertices.keys():
            new_vid += 1
        return new_vid

    @staticmethod
    def calculate_area(vertices: list) -> float:
        """
        Calculate the area of a given polygon using the shoelace formula

        :param vertices: array of [x,y] positions of the ordered vertices in a polygon
        :type vertices: list
        :return: Area of the polygon described by the vertices
        :rtype: float
        """
        x = [i[0] for i in vertices]
        y = [i[1] for i in vertices]
        return abs(0.5 * (np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1))))

    @staticmethod
    def get_distance(v1: vertex.Vertex, v2: vertex.Vertex) -> float:
        """
        Get the distance between two vertices in the system

        :param v1: First vertex
        :type v1: vertex.Vertex
        :param v2: Second vertex
        :type v2: vertex.Vertex
        :return: Distance between the vertices
        :rtype: float
        """
        return np.linalg.norm(np.array([v1.x, v1.y]) - np.array([v2.x, v2.y]))



import numpy as np
import pandas as pd
from typing import Tuple

# TODO Use map() to improve dfs creation

def get_cells_df(frame: object) -> pd.DataFrame:
    """Create a DataFrame with all the information about the cells required
    to calculate the stress tensor.

    :param frame: Frame at which to calculate the stress tensor
    :type frame: object
    :return: Dataframe with ids, centroid position, area and perimeter of cells in the system 
    :rtype: pd.DataFrame
    """
    cells = pd.DataFrame()
    cell_centers_x = [cell.get_cm()[0] for _, cell in frame.cells.items()]
    cell_centers_y = [cell.get_cm()[1] for _, cell in frame.cells.items()]
    cell_ids = [cid for cid, _ in frame.cells.items()]
    cell_areas = [abs(cell.get_area()) for _, cell in frame.cells.items()]
    cell_pressures = [cell.pressure for _, cell in frame.cells.items()]

    cells["ids"] = cell_ids
    cells["xcm"] = cell_centers_x
    cells["ycm"] = cell_centers_y
    cells["area"] = cell_areas
    cells["pressure"] = cell_pressures

    return cells

def get_big_edges_df(frame: object) -> pd.DataFrame:
    """Create a DataFrame with all the information about the edges required
    to calculate the stress tensor.

    :param frame: Frame at which to calculate the stress tensor
    :type frame: object
    :return: Dataframe with ids, edge's stress, and which cells are around it.
    :rtype: pd.DataFrame
    """
    bedges = pd.DataFrame()
    
    bedges_celli = []
    bedges_cellj = []

    bedges_vector = []

    bedges_ids = [big_edge.big_edge_id for _, big_edge in frame.big_edges.items()]
    bedges_stress = [big_edge.tension for _, big_edge in frame.big_edges.items()]
    # bedges_vector= [big_edge.get_vector_from_vertex(big_edge.vertices[0].id) for _, big_edge in frame.big_edges.items()]
    for _, big_edge in frame.big_edges.items():
        bedges_celli.append(big_edge.own_cells[0])
        try:
            bedges_cellj.append(big_edge.own_cells[1])
        except IndexError:
            bedges_cellj.append(-1)

        orientation = np.dot(big_edge.xs, np.roll(big_edge.ys,1)) - np.dot(big_edge.ys, np.roll(big_edge.xs, 1))
        if orientation > 0:
            bedges_vector.append(big_edge.get_vector_from_vertex(big_edge.vertices[0].id))
        else:
            bedges_vector.append(big_edge.get_vector_from_vertex(big_edge.vertices[-1].id))

    bedges["ids"] = bedges_ids
    bedges["stress"] = bedges_stress
    bedges["vector"] = bedges_vector
    bedges["cell1"] = bedges_celli
    bedges["cell2"] = bedges_cellj

    return bedges

def stress_tensor(frame: object, grid: int = 5, radius: float = 1) -> Tuple:
    """Calculate stress tensor in the system using Batchelor's formula.
    Batchelor, G. K. The stress system in a suspension of force-free particles. J. Fluid Mech. 41, 545-570 (1970).

    :param frame: Frame object when to calculate the stress tensor.
    :type frame: object
    :param grid: Amount of bins to divide the space in, defaults to 5
    :type grid: int, optional
    :param radius: Number of average cell radii to perform the summation, defaults to 1
    :type radius: float, optional
    :return: Tuple with stress tensor's values, bin positions and grid centers.
    :rtype: Tuple
    """
    sigmas = {}
    cells = get_cells_df(frame)
    big_edges = get_big_edges_df(frame)

    _, x_bins = np.histogram(cells["xcm"], grid)
    _, y_bins = np.histogram(cells["ycm"], grid)

    x_bin_centers = [(x_bins[ii] + x_bins[ii+1]) / 2 for ii in range(len(x_bins) -1)]
    y_bin_centers = [(y_bins[ii] + y_bins[ii+1]) / 2 for ii in range(len(y_bins) -1)]
    bins_centers = [x_bin_centers, y_bin_centers]

    min_distance = radius * np.sqrt(cells["area"].mean() / np.pi)

    for row in range(grid):
        for column in range(grid):
            center = ((x_bins[row + 1] + x_bins[row]) / 2, (y_bins[column + 1] + y_bins[column]) / 2) 
            current_cell_mesh = cells.loc[(center[0] - cells["xcm"])**2 + (center[1] - cells["ycm"])**2 <= min_distance**2]
            total_area = current_cell_mesh["area"].sum()

            if total_area == 0:
                sigmas[f"{row}{column}"] = np.array([[0, 0], [0, 0]], dtype=float)
                continue

            pressure_area_term = - np.sum([cell["pressure"] * cell["area"] for _, cell in current_cell_mesh.iterrows()])

            current_edges_mesh = big_edges.loc[(big_edges["cell1"]).isin(current_cell_mesh["ids"]) |
                                               (big_edges["cell2"].isin(current_cell_mesh["ids"]))]
            
            tension_xx = 0
            tension_yy = 0
            tension_xy = 0
            for _, bedge in current_edges_mesh.iterrows():
                vector_norm = np.linalg.norm(bedge["vector"])
                tension_xx += bedge["stress"] * (bedge["vector"][0] * bedge["vector"][0]) / vector_norm 
                tension_yy += bedge["stress"] * (bedge["vector"][1] * bedge["vector"][1]) / vector_norm 
                tension_xy += bedge["stress"] * (bedge["vector"][0] * bedge["vector"][1]) / vector_norm 

            sigma_xx = (pressure_area_term + tension_xx) / total_area
            sigma_yy = (pressure_area_term + tension_yy) / total_area
            sigma_xy = tension_xy / total_area

            sigmas[f"{row}{column}"] = np.array([[sigma_xx, sigma_xy], [sigma_xy, sigma_yy]], dtype=float)

    return sigmas, bins_centers, (x_bins, y_bins)


import pytest

import numpy as np
import forsys as fs

@pytest.fixture
def set_up_cell():
    def get_hexagon_vertices(apothem=1):
        hex_vertices = [[apothem * np.cos(ang), apothem * np.sin(ang)] 
                        for ang in np.arange(0, 2*np.pi, 60 * np.pi / 180)]
        return hex_vertices

    vertices = {}
    edges = {}

    for vid, vertex in enumerate(get_hexagon_vertices()):
        vertices[vid] = fs.vertex.Vertex(vid, vertex[0], vertex[1])
        if vid > 0:
            edges[vid - 1] = fs.edge.SmallEdge(vid - 1, vertices[vid - 1], vertices[vid])
    edges[len(edges)] = fs.edge.SmallEdge(len(edges), vertices[0], vertices[vid])


    cell = fs.cell.Cell(0, list(vertices.values()), {})

    
    yield cell


def test_get_area(set_up_cell):
    assert abs(np.round(set_up_cell.get_area(), 6)) == np.round(1.5 * np.sqrt(3), 6)


def test_get_area_sign(set_up_cell):
    assert set_up_cell.get_area_sign() == -1
    set_up_cell.vertices = set_up_cell.vertices[::-1]
    assert set_up_cell.get_area_sign() == 1


def test_get_perimeter(set_up_cell):
    assert int(set_up_cell.get_perimeter()) == 6


import pytest
import numpy as np
from forsys.frames import Frame
from forsys.vertex import Vertex
from forsys.edge import BigEdge, SmallEdge

@pytest.fixture
def set_up_frame():
    vertices = {}
    edges = {}
    cells = {}
    frame = Frame(0, vertices, edges, cells)


    for vid in range(10):
        vertices[vid] = Vertex(vid, vid + 1 , vid + 1)
        if vid > 0:
            edges[vid - 1] = SmallEdge(vid - 1, vertices[vid - 1], vertices[vid])

    frame.big_edges = {
        0: BigEdge(0,
                   vertices=[vertices[0], vertices[1], vertices[2], vertices[3], vertices[4]]),
        1: BigEdge(1, 
                   vertices=[vertices[5],
                             vertices[6],
                             vertices[7],
                             vertices[8],
                             vertices[9]])
    }
    return frame

def test_filter_edges_sg(set_up_frame):
    set_up_frame.filter_edges(method="SG")
    # Check if Savitzky-Golay filter is applied correctly
    assert np.allclose(set_up_frame.big_edges[0].xs, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[0].ys, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[1].xs, [6.0, 7.0, 8.0, 9.0, 10.0])
    assert np.allclose(set_up_frame.big_edges[1].ys, [6.0, 7.0, 8.0, 9.0, 10.0])

def test_filter_edges_none(set_up_frame):
    set_up_frame.filter_edges(method="none")
    # Check if no filtering is applied
    assert np.allclose(set_up_frame.big_edges[0].xs, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[0].ys, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[1].xs, [6.0, 7.0, 8.0, 9.0, 10.0])
    assert np.allclose(set_up_frame.big_edges[1].ys, [6.0, 7.0, 8.0, 9.0, 10.0])

def test_filter_edges_invalid_method(set_up_frame):
    with pytest.raises(ValueError):
        set_up_frame.filter_edges(method="invalid")


class DifferentTissueException(Exception):
    def __init__(self, message='Tissues are too different'):
        super(DifferentTissueException, self).__init__(message)

class SegmentationArtifactException(Exception):
    def __init__(self, message='There might be a segmentation artifact from the TIFF'):
        super(SegmentationArtifactException, self).__init__(message)

class BigEdgesBadlyCreated(Exception):
    def __init__(self, message='Big edge was not created correctly'):
        super(BigEdgesBadlyCreated, self).__init__(message)


import numpy as np
import scipy.spatial as scspace
from copy import deepcopy
import itertools as itert

import forsys.vertex as fvertex
import forsys.edge as fedge
import forsys.cell as fcell

def create_lattice(vertices_voronoi, edges_voronoi, cells_voronoi, **kwargs) -> tuple:
    vertices = {}
    edges = {}
    cells = {}
    for vid, vertex in vertices_voronoi.items():
        vertices[vid] = fvertex.Vertex(vid, vertex[0], vertex[1])
    
    for eid, edge in edges_voronoi.items():
        edges[abs(eid)] = fedge.SmallEdge(abs(eid), vertices[edge[0]], vertices[edge[-1]])

    for cid, cell in cells_voronoi.items():
        vertices_in_cell = []
        for eid in cell:
            which = 0 if eid > 0 else -1
            current_vertex = vertices[edges[abs(eid)].get_vertices_id()[which]]
            vertices_in_cell.append(current_vertex)

        vertices_in_cell = vertices_in_cell[::-1] if cid < 0 else vertices_in_cell
        cells[abs(cid)] = fcell.Cell(abs(cid), vertices_in_cell)

    return vertices, edges, cells

def create_lattice_elements(cell_centers: list, **kwargs) -> tuple:
    """
    Create the vertices, edges and cells from the scipy.spatial.Voronoi() objects

    :param cell_centers: Centroids of the objects 
    :type cell_centers: list
    :return: Vertices, edges and cell's list
    :rtype: tuple
    """
    # if kwargs.get("add_layer", False):
    #     centroid = np.mean(cell_centers)
        
    tessellation = scspace.Voronoi(cell_centers)

    new_vertices = {}
    new_cells = {}
    new_edges = {}
    regions = deepcopy(tessellation.regions)
    big_edge = []

    cnum = 1

    regions = remove_infinite_regions(tessellation, regions, max_distance=kwargs.get("max_distance", 75))

    for c in regions:
        temp_for_cell = []
        temp_vertex_for_cell = []
        if len(c) != 0 and -1 not in c:
            # add first to close the cell_vertices
            c.append(c[0])
            for ii in range(0, len(c) - 1):
                temp_big_edge = []
                x_coordinate = np.around(np.linspace(round(tessellation.vertices[c[ii]][0], 3),
                                                    round(tessellation.vertices[c[ii + 1]][0], 3), 2), 3)
                y_coordinate = np.around(line_eq(tessellation.vertices[c[ii]],
                                                    tessellation.vertices[c[ii + 1]],
                                                    x_coordinate), 3)

                new_edge_vertices = list(zip(x_coordinate, y_coordinate))
                # add new edges to the global list
                for v in range(0, len(new_edge_vertices) - 1):
                    v0 = new_edge_vertices[v]
                    v1 = new_edge_vertices[v + 1]

                    vertex_number_1 = get_vertex_number(v0, new_vertices)

                    vertex_number_2 = get_vertex_number(v1, new_vertices)

                    enum = get_enum([vertex_number_1, vertex_number_2], new_edges)

                    temp_big_edge.append(enum)
                    temp_for_cell.append(enum)
                    temp_vertex_for_cell.append(vertex_number_1)
                    temp_vertex_for_cell.append(vertex_number_2)

                big_edge.append(temp_big_edge)

            area_sign = get_cell_area_sign(temp_vertex_for_cell, new_vertices)
            new_cells[-1 * cnum * area_sign] = temp_for_cell
            cnum += 1

    return new_vertices, new_edges, new_cells


def remove_infinite_regions(tessellation, regions: list, max_distance: float = 50) -> list:
    """
    Remove regions that have a vertex too far away to be part of the tissue. This solves vertices placed in
    'infinity' by the tessellation

    :param regions: List of the voronoi regions
    :type regions: list
    :param max_distance: Maximum distance of a vertex in a polygon
    :type max_distance: float, optional
    :return: List of regions without the offending ones
    :rtype: list
    """
    to_delete = []
    used_vertices = []
    for c in regions:
        if len(c) != 0 and -1 not in c:
            polygon_vertices = [list(tessellation.vertices[vid]) for vid in c]
            matrix = distance_matrix(polygon_vertices)
            if np.max(matrix) > max_distance:
                to_delete.append(c)
    for c in to_delete:
        regions.remove(c)
    return regions


def line_eq(p0: float, p1: float, x: np.ndarray) -> list:
        """
        Get the value of linear function that goes through p0 and p1 at x

        :param p0: First point to join
        :type p0: float
        :param p1: Second point to join
        :type p0: float
        :param x: Independent variable of the equation
        :type p0: list
        :return: Value of y(x) for a line that goes through p0 and p1
        """
        p0 = np.around(p0, 3)
        p1 = np.around(p1, 3)
        m = (p1[1] - p0[1]) / (p1[0] - p0[0])
        return p0[1] + m * (x - p0[0])


def get_cell_area_sign(cell: list, all_vertices: dict) -> int:
    """
    Get the orientation of the polygon for a cell_vertices using the sign of the area through gauss formula

    :param cell: Vertices of the cell_vertices to determine orientation
    :type cell: list
    :param all_vertices: Dictionary with all the vertices in the lattice
    :type all_vertices: dict
    :return: Sign of the cell_vertices area to determine orientation of the polygon
    :rtype: int
    """
    return int(np.sign(get_cell_area(cell, all_vertices)))


def get_cell_area(cell_vertices: list, vertices: dict) -> float:
    """
    Area of a polygon using the shoelace formula

    :param cell_vertices: List of vertex's id of the polygon
    :param vertices: Dictionary of vertices in the system
    :return: Area of the polygon. Positive number indicates clockwise orientation, negative number \
        counterclockwise.
    """
    x = [vertices[i][0] for i in cell_vertices]
    y = [vertices[i][1] for i in cell_vertices]
    return 0.5 * (np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def get_vertex_number(vertex: list, vertices: dict) -> int:
    """
    Get id of the vertex from the vertex position or new possible ID if the vertex is already known

    :param vertex: Vertex position
    :type vertex: list
    :param vertices: Dictionary of the vertices currently identified
    :type vertices: dict
    :return: Vertex id if the vertex exists, if not the next possible id to assign
    :rtype: int
    """
    if vertex in vertices.values():
        vertex_number = list(vertices.keys())[list(vertices.values()).index(vertex)]
    else:
        if len(vertices) > 0:
            vertex_number = max(vertices.keys()) + 1
        else:
            vertex_number = 1
        vertices[vertex_number] = vertex
    return vertex_number

def get_enum(edge: list, edges: dict) -> int:
    """
    Get id of the edge from the vertex position or new possible ID if the edge is already known

    :param edge: id of the vertices that create the edge
    :type edge: list
    :param edges: Dictionary of the edges currently identified
    :type edges: dict
    :return: edge id if the edge exists, if not the next possible id to assign
    :rtype: int
    """
    if edge in edges.values():
        enum = list(edges.keys())[list(edges.values()).index(edge)]
    elif edge[::-1] in edges.values():
        enum = - list(edges.keys())[list(edges.values()).index(edge[::-1])]
    else:
        if len(edges) > 0:
            enum = max(edges.keys()) + 1
        else:
            enum = 1
        edges[enum] = [edge[0], edge[1]]
    return enum


def add_voronoi_centers(cell_centers):
    matrix = distance_matrix(cell_centers)
    ave_distance = np.median(matrix) / 10
    max_x = np.max(list(zip(*cell_centers))[0])
    min_x = np.min(list(zip(*cell_centers))[0])
    max_y = np.max(list(zip(*cell_centers))[1])
    min_y = np.min(list(zip(*cell_centers))[1])

    range_x = np.linspace(min_x - ave_distance, max_x + ave_distance, 10)
    range_y = np.linspace(min_y - ave_distance, max_y + ave_distance, 10)
    new_cell_centers = []
    for ii in range_x:
        new_cell_centers.append((ii, min_y - ave_distance))
        new_cell_centers.append((ii, max_y + ave_distance))
    for kk in range_y:
        new_cell_centers.append((min_x - ave_distance, kk))
        new_cell_centers.append((max_x + ave_distance, kk))
    
    return new_cell_centers


def distance_matrix(vertices: list) -> list:
    matrix = np.zeros((len(vertices), len(vertices)))
    for (element_1, element_2) in itert.product(vertices, repeat=2):
        ii = vertices.index(element_1)
        jj = vertices.index(element_2)
        distance_element = np.array(element_1) - np.array(element_2)
        matrix[jj, ii] = np.linalg.norm(distance_element)
    return matrix


import pytest
import os 

import numpy as np
import forsys as fs

import scipy.stats as sps


@pytest.fixture
def furrow():
    frames = {}
    for ii in range(0, 8):
        surfaceEvolver = fs.surface_evolver.SurfaceEvolver(os.path.join(
                                                            "tests", 
                                                            "data", 
                                                            "furrow_gauss_velocity", 
                                                            f"stage{ii}.dmp"))
        frames[ii] = fs.frames.Frame(ii,
                                    surfaceEvolver.vertices,
                                    surfaceEvolver.edges, 
                                    surfaceEvolver.cells, 
                                    time=ii, 
                                    gt=True)
    forsys = fs.ForSys(frames, cm=False)
    yield forsys


def test_pressure_initial_furrow(furrow):
    furrow.build_force_matrix(when=0)
    furrow.solve_stress(when=0)

    furrow.build_pressure_matrix(when=0)
    furrow.solve_pressure(when=0, method="lagrange_pressure")

    pressures_df = furrow.frames[0].get_pressures()
    pressures_df["gt_pressure"] = (pressures_df["gt_pressure"] - pressures_df["gt_pressure"].mean()) / pressures_df["gt_pressure"].std()
    pressures_df["pressure"] = (pressures_df["pressure"] - pressures_df["pressure"].mean()) / pressures_df["pressure"].std()
    
    correlation = sps.pearsonr(pressures_df['gt_pressure'], pressures_df['pressure'])[0]

    assert correlation > 0.99


def test_pressure_last_furrow(furrow):
    last_frame = len(furrow.frames) - 1
    furrow.build_force_matrix(when=last_frame, circle_fit_method="dlite")
    furrow.solve_stress(when=last_frame)

    furrow.build_pressure_matrix(when=last_frame)
    furrow.solve_pressure(when=last_frame, method="lagrange_pressure")

    pressures_df = furrow.frames[last_frame].get_pressures()
    pressures_df["gt_pressure"] = (pressures_df["gt_pressure"] - pressures_df["gt_pressure"].mean()) / pressures_df["gt_pressure"].std()
    pressures_df["pressure"] = (pressures_df["pressure"] - pressures_df["pressure"].mean()) / pressures_df["pressure"].std()
    
    correlation = sps.pearsonr(pressures_df['gt_pressure'], pressures_df['pressure'])[0]

    assert correlation > 0.99


def test_pressure_all_frames_furrow(furrow):
    for ii in range(0, 8):
        furrow.build_force_matrix(when=ii, circle_fit_method="dlite")
        furrow.solve_stress(when=ii)

        furrow.build_pressure_matrix(when=ii)
        furrow.solve_pressure(when=ii, method="lagrange_pressure")

        pressures_df = furrow.frames[ii].get_pressures()
        pressures_df["gt_pressure"] = (pressures_df["gt_pressure"] - pressures_df["gt_pressure"].mean()) / pressures_df["gt_pressure"].std()
        pressures_df["pressure"] = (pressures_df["pressure"] - pressures_df["pressure"].mean()) / pressures_df["pressure"].std()
        
        correlation = sps.pearsonr(pressures_df['gt_pressure'], pressures_df['pressure'])[0]
        
        assert correlation > 0.99


import numpy as np

def get_versors_area(vertices, edges, cells, vid, metadata):
    try:
        if metadata['area_target'] == 'min':
            areaTarget = get_minmax_area(cells)[0]
        elif metadata['area_target'] == 'max':
            areaTarget = get_minmax_area(cells)[1]
        elif metadata['area_target'] == 'ave':
            areaTarget = get_minmax_area(cells)[2]       
        else:
            areaTarget = metadata['area_target']
    except KeyError:
        print("***WARNING**** No area target defined, using average")
        areaTarget = get_minmax_area(cells)[2]
    
    areax = 0
    areay = 0
    print("Calculating area terms... vID: ", vid, "with area target: ", areaTarget)

    for cellID in vertices[vid].ownCells:
        cell = cells[cellID]
        area = abs(cell.get_area())
        deltay = cell.get_next_vertex(vertices[vid]).y - cell.get_previous_vertex(vertices[vid]).y
        deltax = cell.get_next_vertex(vertices[vid]).x - cell.get_previous_vertex(vertices[vid]).x
        # print("Deltas: ", deltax, deltay)
        # print("Deltas: ", deltax, deltay)
        # print("areas", area, areaTarget, area-areaTarget)
        areax -= (area-areaTarget) * deltay
        areay += (area-areaTarget) * deltax
   
    return areax, areay

def get_versors_perimeter(vertices, edges, cells, vid, metadata):
    try:
        if metadata['perimeter_target'] == 'min':
            perimeterTarget = get_minmax_perimeter(cells)[0]
        elif metadata['perimeter_target'] == 'max':
            perimeterTarget = get_minmax_perimeter(cells)[1]
        elif metadata['perimeter_target'] == 'ave':
            perimeterTarget = get_minmax_perimeter(cells)[2]
        else:
            perimeterTarget = metadata['perimeter_target']
    except KeyError:
        print("No area target defined, using average")
        perimeterTarget = get_minmax_perimeter(cells)[2]

    print("Calculating perimeter terms... vID: ", vid, "with perimeter target: ", perimeterTarget)
    perimeterx = 0
    perimetery = 0
    for cellID in vertices[vid].ownCells:
        cell = cells[cellID]
        perimeter = cell.get_perimeter()
        deltaxPrev = vertices[vid].x - cell.get_previous_vertex(vertices[vid]).x
        deltaxNext = vertices[vid].x - cell.get_next_vertex(vertices[vid]).x
        deltayPrev = vertices[vid].y - cell.get_previous_vertex(vertices[vid]).y
        deltayNext = vertices[vid].y - cell.get_next_vertex(vertices[vid]).y

        print("Deltas: ", deltaxPrev, deltaxNext)
        print("Deltas: ", deltaxNext, deltayNext)
        print("areas", perimeter, perimeterTarget, perimeter-perimeterTarget)
        print("----------------------------------------------------------------------------------------")

        modPrev = np.sqrt(deltaxPrev**2+deltayPrev**2)
        modNext = np.sqrt(deltaxNext**2+deltayNext**2)
        perimeterx += -2*(perimeter-perimeterTarget)*(deltaxPrev/modPrev+deltaxNext/modNext)
        perimetery += -2*(perimeter-perimeterTarget)*(deltayPrev/modPrev+deltayNext/modNext)

    return perimeterx, perimetery


def list_areas(cells):
    areas = []
    for _, cell in cells.items():
        areas.append(cell.get_area())
    return areas

def list_perimeters(cells):
    perim = []
    for _, cell in cells.items():
        perim.append(cell.get_perimeter())
    return perim

def get_minmax_area(cells):
    areas = list_areas(cells)
    return np.min(np.abs(areas)), np.max(np.abs(areas)), np.mean(np.abs(areas))

def get_minmax_perimeter(cells):
    perims = list_perimeters(cells)
    return min(perims), max(perims), np.mean(perims)


from dataclasses import dataclass
import numpy as np
from typing import Tuple
import scipy.optimize as sco
import circle_fit as cfit

import forsys.virtual_edges as ve

@dataclass 
class Cell:
    """
    Class representation of the cells objects.

    :param id: Unique identifier of the cell
    :type id: int
    :param vertices: Vertices that form the cell
    :type vertices: list
    :param is_border: True if the cell is in the border , defaults to False
    :type is_border: float
    :param gt_pressure: Ground truth pressure of the cell, defaults to None
    :type gt_pressure: float
    :param pressure: Inferred pressure of the cell, defaults to None
    :type pressure: float
    """
    id: int
    vertices: list
    is_border: bool = False
    
    gt_pressure: float = None
    pressure: float = None
    center_method: str = "dlite"


    def __post_init__(self):
        """Constructor method
        """
        self.center_x = None
        self.center_y = None
        self.neighbors = None

        for v in self.vertices:
            v.add_cell(self.id)

        self.center_x, self.center_y = ve.calculate_circle_center(vertices=self.vertices, method=self.center_method)
        
    def __del__(self):
        for v in self.vertices:
            v.remove_cell(self.id)
        
    def get_cell_vertices(self) -> list:
        """
        Get the list of vertices corresponding to the cell

        :return: List of the cell's vertices
        :rtype: list
        """
        return self.vertices

    def get_next_vertex(self, v: object) -> object:
        """
        Get the next vertex from a given one in the cell ordering

        :param v: Current vertex after which the next is calculated
        :type cid: object
        :return: Object reference for the next vertex in the ordering
        :rtype: object
        """
        return self.vertices[(self.vertices.index(v) + self.get_area_sign()) % len(self.vertices)]
    
    def get_previous_vertex(self, v: object) -> object:
        """
        Get the previous vertex from a given one in the cell ordering

        :param v: Current vertex after which the previous is calculated
        :type cid: object
        :return: Object reference for the previous vertex in the ordering
        :rtype: object
        """
        return self.vertices[(self.vertices.index(v) - self.get_area_sign())% len(self.vertices)]
        
    def get_cm(self) -> list:
        """
        Get the centroid of the cell

        :return:[x, y] list with the centroid position
        :rtype: list
        """
        return [np.mean([v.x for v in self.vertices]), np.mean([v.y for v in self.vertices])]

    def get_area_sign(self) -> int:
        """
        Get the sign of the area, after the shoelace algorithm

        :return: -1 means the cell is ordered clockwise, +1 counterclockwise.
        :rtype: int
        """
        return int(np.sign(self.get_area()))

    def get_area(self) -> float:
        """
        Get the area of cell, using the shoelace formula

        :return: Cell's area
        :rtype: float
        """
        x = [i.get_coords()[0] for i in self.vertices]
        y = [i.get_coords()[1] for i in self.vertices]
        return 0.5 * (np.dot(x, np.roll(y,1)) - np.dot(y, np.roll(x, 1)))

    def get_perimeter(self) -> float:
        """
        Get the perimeter of the cell

        :return: Cell's perimeter
        :rtype: float
        """
        perimeter = 0
        for i in self.vertices:
            diffx = i.x - self.get_next_vertex(i).x
            diffy = i.y - self.get_next_vertex(i).y
            perimeter +=  np.sqrt(diffx**2 + diffy**2)
        return perimeter

    def replace_vertex(self, vold: object, vnew: object) -> None:
        """
        Replace vertex vold with vnew in the vertices list and
        add this cell to that vertex. If the vertex is already
        in the cell, remove it.

        :param vold: Old vertex object to replace
        :type vold: object
        :param vnew: New vertex to replace in vold positions
        :type vnew: object
        """
        vertices_ids = [v.id for v in self.vertices]
        if vnew.id in vertices_ids:
            self.vertices.remove(vold)
        else:
            self.vertices[vertices_ids.index(vold.id)] = vnew
            # add cell to vertex
            vnew.add_cell(self.id)


    def calculate_neighbors(self) -> list:
        """
        Get all neighboring cell's ids and set the corresponding class variable

        :return: IDs of neighboring cells
        :rtype: list
        """
        # TODO Functionalize
        current_cells = set()
        for vertex in self.vertices:
            [current_cells.add(cell_id) for cell_id in vertex.ownCells]
        current_cells = list(current_cells)
        
        current_cells.remove(self.id)
        self.neighbors = current_cells
        return self.neighbors
    

    def get_edges(self):
        cell_edges = []
        for vnum in range(0, len(self.vertices) - 1):
            local_edges = list(set(self.vertices[vnum].ownEdges) &
                               set(self.vertices[vnum + 1].ownEdges))
            cell_edges.append(local_edges[0])
        return cell_edges


import pytest
import os 

import numpy as np
import forsys as fs

@pytest.fixture
def extra_cell():
    frames = {}
    skeleton = fs.skeleton.Skeleton(os.path.join("tests", 
                                                 "data", 
                                                 "test_nonzero.tif"))
    vertices, edges, cells = skeleton.create_lattice()
    vertices, edges, cells, _ = fs.virtual_edges.generate_mesh(vertices, 
                                                                edges, 
                                                                cells, 
                                                                ne=6)
    frames[0] = fs.frames.Frame(0,
                                vertices,
                                edges, 
                                cells, 
                                time=0)
    forsys = fs.ForSys(frames, cm=False)

    yield forsys

def test_fmatrix_three_non_zero(extra_cell):
    extra_cell.build_force_matrix()
    extra_cell.solve_stress()

    assert not np.any(extra_cell.frames[0].forces == 0)


import numpy as np
import pandas as pd
import scipy.signal as spsignal
import os

from dataclasses import dataclass, field
from warnings import warn

import forsys as fs
import forsys.edge as fedge

@dataclass
class Frame():
    """
    Class representation of a microscopy frame.

    :param frame_id: Unique identifier of the frame
    :type frame_id: int
    :param vertices: Vertices dictionary for the whole frame
    :type vertices: dict
    :param edges: SmallEdges dictionary for the whole frame
    :type edges: dict
    :param cells: Cells dictionary for the whole frame
    :type cells: dict
    :param time: Real time corresponding to the frame. This is used to calcualte dynamical information.
    :type time: float
    :param gt: If True, the BigEdges are assigned their ground truths. Defaults to False
    :type gt: bool, optional
    """
    frame_id: int
    vertices: dict
    edges: dict
    cells: dict

    time: float=0.0

    gt: bool=False

    def __post_init__(self):
        """Constructor method
        """
        self.big_edge_gt_tension = {}
        self.big_edge_tension = {}
        self.big_edges = {}

        self.big_edges_list = fs.virtual_edges.create_edges_new(self.vertices, self.cells)
        # create big edge objects
        for big_edge_id, big_edge in enumerate(self.big_edges_list):
            vertices_objects = [self.vertices[vid] for vid in big_edge]
            self.big_edges[big_edge_id] = fedge.BigEdge(big_edge_id, vertices_objects)

        self.external_edges_id = [self.big_edges_list.index(e) 
                                    for e in fs.virtual_edges.get_border_edge(self.big_edges_list, 
                                                                self.vertices)]
        
        self.internal_big_edges_vertices = [edge for eid, edge in enumerate(self.big_edges_list) 
                                        if eid not in self.external_edges_id and 
                                        (len(self.vertices[edge[0]].ownCells) > 2 or  
                                        len(self.vertices[edge[-1]].ownCells) > 2)]
        self.internal_big_edges = [self.big_edges[eid] for eid, edge in enumerate(self.big_edges_list) 
                                    if eid not in self.external_edges_id and 
                                    (len(self.vertices[edge[0]].ownCells) > 2 or  
                                    len(self.vertices[edge[-1]].ownCells) > 2)]
        
        for _, cell in self.cells.items():
            cell.calculate_neighbors()

        self.border_vertices = self.get_external_edges()
        if self.gt:
            for big_edge_id, big_edge in self.big_edges.items():
                objects = [self.edges[eid].gt for eid in big_edge.edges]
                big_edge.gt = np.mean(objects)

    def assign_tensions(self, xresult: list) -> None:
        """Assign tensions to the small edges following a list of values. 
        Deprecated function.

        :param xresult: Inference result from the matrix solver
        :type xresult: list
        """
        warn("Legacy function, soon to be deprecated", DeprecationWarning, 2)
        for i in range(0, len(self.big_edges_list)):
            e = self.big_edges_list[i]
            edges_to_use = self.get_big_edge_edgesid(e)
            for e in edges_to_use:
                self.edges[e].tension = xresult[i]

    def get_big_edges(self, use_all: bool = False) -> list:
        """Get list of big edges in the frame, depending on whether
        to use externals or not

        :param use_all: If True external edges are also used, defaults to False
        :type use_all: bool, optional
        :return: List of big edges to use
        :rtype: list
        """
        if use_all:
            big_edges_to_use = self.big_edges       
        else:
            big_edges_to_use = [big_edge for big_edge in self.big_edges.values() 
                                if not big_edge.external]
        return big_edges_to_use
        
    def get_external_edges(self) -> list:
        """Get list of external edges in the system

        :return: List of external edges
        :rtype: list
        """
        return [big_edge for big_edge in self.big_edges.values() 
                                if big_edge.external]
    
    def get_external_edges_ids(self) -> list:
        """Get list of the external's edges IDs.

        :return: List of IDs for all external big edges in the system.
        :rtype: list
        """
        return [big_edge.big_edge_id for big_edge in self.big_edges.values() 
                                if big_edge.external]

    def assign_gt_tensions_to_big_edges(self, gt_tensions: dict, use_all: bool = False) -> None:
        """Assign the ground truth stress to the big edges from a dictionary.

        :param gt_tensions: Dictionary with the ground truth stress values 
        :type gt_tensions: dict
        :param use_all: If True external big edges are also included, defaults to False
        :type use_all: bool, optional
        """
        if use_all:
            big_edges_to_use = self.get_big_edges(use_all=True)
        else:
            big_edges_to_use = self.get_big_edges(use_all=False)
        
        for big_edge_id, big_edge in enumerate(big_edges_to_use):
            big_edge.gt = gt_tensions[big_edge_id]
            for edge_id in big_edge.edges:
                self.edges[edge_id].gt = big_edge.gt

    def assign_tensions_to_big_edges(self) -> None:
        """Assign the inferred tension to the big edge. 
        """
        for big_edge_id, big_edge in self.big_edges.items():
            objects = [self.edges[eid].tension for eid in big_edge.edges]
            self.big_edges[big_edge_id].tension = np.mean(objects)


    def assign_gt_small_edges(self, big_edges_to_use:str = "internal") -> None:
        """Assign the ground truth stress values to the small edges.

        :param big_edges_to_use: "ext" includes external edges, defaults to "internal"
        :type big_edges_to_use: str, optional
        """
        if big_edges_to_use == "ext":
            big_edges_to_use = self.big_edges_list
        else:
            big_edges_to_use = self.internal_big_edges

        for big_edge in big_edges_to_use:
            for small_edge_id in big_edge.edges:
                self.edges[small_edge_id].gt = big_edge.gt

    def assign_pressures(self, pressures: dict, mapping: dict) -> None:
        """Assign the ground truth pressures to the cells

        :param pressures: Inferred pressures dictionary
        :type pressures: dict
        :param mapping: Mapping between cells' ids and IDs in the pressure dictionary
        :type mapping: dict
        """
        for cid, cell in self.cells.items():
            key_to_use = mapping[cid]
            cell.pressure = pressures[key_to_use]

    def get_big_edge_edgesid(self, big_edge_vertices: list) -> list:
        """Get ids of the vertices that form big edges

        :param big_edge_vertices: List of vertices in the big edge
        :type big_edge_vertices: list
        :return: List of ids
        :rtype: list
        """
        # TODO legacy function to-remove
        warn("Legacy function, soon to be deprecated", DeprecationWarning, 2)
        return [list(set(self.vertices[big_edge_vertices[vid]].ownEdges) & 
                set(self.vertices[big_edge_vertices[vid+1]].ownEdges))[0]
                for vid in range(0, len(big_edge_vertices)-1)]

    def get_tensions(self, with_border: bool = False) -> pd.DataFrame:
        """Create tension DataFrame for the ground truth and inferred values.

        :param with_border: True includes the values for the external edges, defaults to False
        :type with_border: bool, optional
        :return: DataFrame with the ID of the edge, its groun truth stress and the inferred one
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame.from_dict({beid: big_edge.gt for beid, big_edge in self.big_edges.items()}.items()).rename(columns={0: 'id', 1: 'gt'})
        df['stress'] = [big_edge.tension for big_edge in self.big_edges.values()]
        if not with_border:
            # Only return results that don't belong the edges in the border
            df = df.loc[~df.id.isin(self.get_external_edges_ids())]

        return df
    

    def get_gt_tensions(self, with_border: bool = False) -> pd.DataFrame:
        df = pd.DataFrame.from_dict({beid: big_edge.gt for beid, big_edge in self.big_edges.items()}.items()).rename(columns={0: 'id', 
                                                                                                                                1: 'gt'})
        if not with_border:
            df = df.loc[~df.id.isin(self.get_external_edges_ids())]
        return df


    def get_pressures(self) -> pd.DataFrame:
        """Create the pressures DataFrame.

        :return: DataFrame with the ID of the cells, their ground truth pressure and the inferred value
        :rtype: pd.DataFrame
        """
        pressure_df = pd.DataFrame.from_dict({cid: cell.gt_pressure for cid, cell in self.cells.items()}.items()).rename(columns={0: "id", 1: "gt_pressure"})
        inferred_pressures = [cell.pressure for cell in self.cells.values()]
        pressure_df["pressure"] = np.array(inferred_pressures).astype(float)

        return pressure_df

    def export_tensions(self, fname: str, folder: str = "", is_gt: bool = True, with_border: bool = False) -> str:
        """Create DataFrame from dictionary of either the ground truth value
        or the inferred ones.

        :param fname: File name to save csv file
        :type fname: str
        :param folder: Folder to save output to, defaults to ""
        :type folder: str, optional
        :param is_gt: True saves the ground truth values, False saves the inferred ones, defaults to True
        :type is_gt: bool, optional
        :param with_border: True saves values including external edges, defaults to False
        :type with_border: bool, optional
        :return: Result of pd.DataFrame().to_csv() function call
        :rtype: str
        """
        if is_gt:
            df = pd.DataFrame.from_dict(self.big_edge_gt_tension.items()).rename(columns={0: 'id', 1: 'tension'})
        else:
            df = pd.DataFrame.from_dict(self.big_edge_tension.items()).rename(columns={0: 'id', 1: 'tension'})
        


        if not with_border:
            # Only return results that don't belong the edges in the border
            df = df.loc[~df.id.isin(self.get_external_edges_ids())]
            
        return df.to_csv(os.path.join(folder, fname), index=False)

    def get_big_edge_by_cells(self, c1_id: int, c2_id: int) -> fedge.BigEdge:
        """Get big edge object from the cells that compose it

        :param c1_id: ID of one the cells
        :type c1_id: int
        :param c2_id: ID of the other cell
        :type c2_id: int
        :return: BigEdge object that is the interface between both cells
        :rtype: fedge.BigEdge
        """
        cell_1 = self.cells[c1_id]
        cell_2 = self.cells[c2_id]
    
        c1_vertices = [v.id for v in cell_1.vertices if len(v.ownEdges) < 3]
        c2_vertices = [v.id for v in cell_2.vertices if len(v.ownEdges) < 3]
        vertices_in_common = np.intersect1d(c1_vertices, 
                                            c2_vertices)
        shared_edge = list(set([edge for vid in vertices_in_common for edge in self.vertices[vid].own_big_edges]))
        return self.big_edges[shared_edge[0]]

    def calculate_stress_tensor(self, coarsing: int = 5, radius: float = 1) -> None:
        """Calculate the principal components of the stress tensor.

        :param coarsing: Number of bins to divide the space, defaults to 5
        :type coarsing: int, optional
        :param radius: Number of average cell raii to integrate the tensor, defaults to 1
        :type radius: float, optional
        """
        self.stress_tensor = fs.stress_tensor.stress_tensor(self, coarsing, radius)

        self.principal_stress = {}

        for row in range(len(self.stress_tensor[1][0])):
            for column in range(len(self.stress_tensor[1][1])):
                principal_component = np.linalg.eig(self.stress_tensor[0][f"{row}{column}"])
                self.principal_stress[(self.stress_tensor[1][0][row], 
                                            self.stress_tensor[1][1][column])] = principal_component
        
    def filter_edges(self, method: str = "SG") -> None:
        """
        Filter the edges of the frame using the specified method.

        :param method: The filtering method to use. Options are "SG" for Savitzky-Golay filter or "none" for no filtering.
        :type method: str, optional
        :raises ValueError: If the specified method is not available.
        :return: None
        """
        if method == "SG":
            filtering_function = spsignal.savgol_filter
            arguments = (5, 3)
        elif method == "none":
            filtering_function = lambda x, *args: x
            arguments = ()
        else:
            raise ValueError("Method not available")

        for index, big_edge in self.big_edges.items():
            try:
                new_xs = filtering_function(big_edge.xs, *arguments)
                new_ys = filtering_function(big_edge.ys, *arguments)
            except ValueError:
                new_xs = big_edge.xs
                new_ys = big_edge.ys
            for index, vertex in enumerate(big_edge.vertices):
                vertex.x = new_xs[index]
                vertex.y = new_ys[index]

    def get_cell_properties_df(self, center_method) -> pd.DataFrame:
        all_ids = []
        all_centerx = []
        all_centery = []
        all_perimeters = []
        all_areas = []
        all_pressures = []

        for cell in self.cells.values():
            center = fs.virtual_edges.calculate_circle_center(cell.vertices, center_method)
            all_ids.append(cell.id)
            all_centerx.append(center[0])
            all_centery.append(center[1])
            all_perimeters.append(cell.get_perimeter())
            all_areas.append(cell.get_area())
            all_pressures.append(cell.pressure)

        return pd.DataFrame({"cid": all_ids,
                             "centerx": all_centerx,
                             "centery": all_centery,
                             "perimeter": all_perimeters,
                             "area": all_areas,
                             "pressure": all_pressures})


    def get_edges_props_df(self) -> pd.DataFrame:
        all_ids = []
        all_tensions = []
        all_gt_tensions = []
        all_posx = []
        all_posy = []

        for edge in self.big_edges.values():
            all_ids.append(edge.big_edge_id)
            all_tensions.append(edge.tension)
            all_gt_tensions.append(edge.gt)
            all_posx.append(np.mean(edge.xs))
            all_posy.append(np.mean(edge.ys))

        return pd.DataFrame({"eid": all_ids,
                             "posx": all_posx,
                             "posy": all_posy,
                             "tension": all_tensions,
                             "gt_tension": all_gt_tensions})


import pytest
import os 
import pandas as pd
import numpy as np
import forsys as fs

from sklearn.metrics import mean_absolute_percentage_error, r2_score

@pytest.fixture
def experiment_1():
    frames = {}
    skeleton = fs.skeleton.Skeleton(os.path.join("tests", 
                                                 "data", 
                                                 "experimental", 
                                                 f"exp_1.tif"), mirror_y=False)
    vertices, edges, cells = skeleton.create_lattice()
    vertices, edges, cells, _ = fs.virtual_edges.generate_mesh(vertices, edges, cells, ne=6)
    frames[0] = fs.frames.Frame(0, vertices, edges, cells, time=0)
    frames[0].filter_edges(method="SG")
    forsys = fs.ForSys(frames, cm=False, initial_guess={})

    df = pd.read_csv(os.path.join("tests", "data", "experimental", "exp_1.csv"))
    
    yield (forsys, df)

def test_myosin_correlation(experiment_1):
    experiment = experiment_1[0]
    df = experiment_1[1]

    experiment.build_force_matrix(when=0,
                            metadata={"ignore_four": False},
                            angle_limit=np.inf,
                            circle_fit_method="dlite")
    experiment.solve_stress(when=0, 
                            b_matrix="static",
                            velocity_normalization=0,
                            adimensional_velocity=False,
                            method="lsq",
                            use_std=False,
                            allow_negatives=False)

    experiment.build_force_matrix(when=0)
    experiment.solve_stress(when=0, allow_negatives=False)

    tensions_df = experiment.frames[0].get_tensions()
    rounded_tensions = tensions_df["stress"].round(2)
    rounded_df = df["stress"].round(2)
    assert np.all(rounded_tensions.values == rounded_df.values)


from PIL import Image
import numpy as np
import itertools
import math
from scipy import interpolate

from forsys.vertex import Vertex 
from forsys.edge import SmallEdge, BigEdge 


def read_myosin(frame,
                tiff_path,
                integrate=False,
                normalize="average",
                corrected_by_tilt=False,
                layers: int = 1,
                **kwargs):
    image = Image.open(tiff_path)

    if kwargs.get("use_all", False):
        big_edges_to_use = frame.big_edges_list
    else:
        big_edges_to_use = frame.internal_big_edges

    return get_intensities(big_edges_to_use,
                           image,
                           integrate,
                           normalize,
                           corrected_by_tilt,
                           layers,
                           **kwargs)


def get_intensities(big_edges,
                    image,
                    integrate=False,
                    normalize="average",
                    corrected_by_tilt=False,
                    layers: int = 1,
                    **kwargs):
    intensities = {}

    if corrected_by_tilt:
        image_to_use = Image.fromarray(modified_intensities(image))
    else:
        image_to_use = image

    for be_id, big_edge in enumerate(big_edges):
        intensities_per_edge = []
        if integrate:
            vertices, length = get_interpolation(big_edge, layers, **kwargs)
            intensity_to_use = sum(map(image_to_use.getpixel,
                                       vertices)) / length
        else:
            for vertex in big_edge.vertices:
                intensity = get_intensity(image_to_use, vertex, layers, **kwargs)
                intensities_per_edge.append(intensity)

            intensity_to_use = np.mean(list(map(np.median,
                                                intensities_per_edge)))
        try:
            key_to_use = big_edges.index(big_edge)
        except ValueError:
            key_to_use = "ext_"+str(be_id)

        intensities[key_to_use] = intensity_to_use
        # big_edge.gt = intensities[key_to_use]

    intensities_only_internal = {k: v for k, v in intensities.items() if not isinstance(k, str)}

    if normalize == "maximum":
        raise NotImplementedError
    elif normalize == "average":
        mean_value = np.mean(list(intensities_only_internal.values()))
        intensities_only_internal = {k: v/mean_value for k, v in intensities_only_internal.items()}

    
    for be_id, big_edge in enumerate(big_edges):
        key_to_use = big_edges.index(big_edge)
        big_edge.gt = intensities_only_internal[key_to_use]

    return intensities_only_internal


# def rotate_big_edge(big_edge_prev, angle):
#     Rmatrix = np.array([[np.cos(angle), - np.sin(angle)],
#                        [np.sin(angle), - np.cos(angle)]])
    
#     vertices = {}
#     for vid in range(len(big_edge_prev.xs)):
#         new_x, new_y = np.array(np.matmul(Rmatrix, [big_edge_prev.xs[vid], big_edge_prev.ys[vid]]))
#         vertices[10000 + vid] = Vertex(10000 + vid, new_x, new_y)
#         if vid > 0:
#             edges[10000 + vid] = SmallEdge(vid - 1, vertices[vid - 1], vertices[vid])

#     big_edge = BigEdge(0, list(vertices.values()))

#     return big_edge


def get_intensity(image, vertex, layers=1, **kwargs):
    offset = kwargs.get("offset", [0, 0])
    rescale = kwargs.get("rescale", [1, 1])
    x_y_position = [(vertex.x * rescale[0]) + offset[0], (vertex.y * rescale[1]) + offset[1]]
    pixel_positions = get_layer_elements(x_y_position, layers)
    return list(map(image.getpixel, pixel_positions))


def get_layer_elements(position, layers=1):
    pixel_positions = []
    layer_range = np.arange(-layers, layers + 1)
    for (ii, kk) in itertools.product(layer_range, layer_range):
        xy_pixel = (position[0] + ii, position[1] + kk)
        pixel_positions.append(xy_pixel)
    return pixel_positions


def get_interpolation(big_edge, layers, **kwargs):
    offset = kwargs.get("offset", [0, 0])
    rescale = kwargs.get("rescale", [1, 1])
    length = 0
    all_vertices = set()
    xs_values = [(value * rescale[0]) + offset[0] for value in big_edge.xs]
    ys_values = [(value * rescale[1]) + offset[1] for value in big_edge.ys]
    # xy_pairs = list(zip(big_edge.xs, big_edge.ys))
    xy_pairs = list(zip(xs_values, ys_values))

    for ii in range(1, len(xy_pairs)):
        current_vertex = list(map(math.ceil, xy_pairs[ii - 1]))
        next_vertex = list(map(math.ceil, xy_pairs[ii]))
        all_vertices.update(walk_two_vertices(current_vertex, next_vertex, layers))
        length += np.linalg.norm(np.array(xy_pairs[ii - 1]) - np.array(xy_pairs[ii]))

    return all_vertices, length


def walk_two_vertices(v0, v1, layers):
    vertices_to_return = set()
    diff_x = abs(v0[0] - v1[0])
    diff_y = abs(v0[1] - v1[1])
    axis = 0 if diff_x > diff_y else 1
    delta = 1 if v0[axis] < v1[axis] else -1
    interpolation = interpolate.interp1d([v0[axis], v1[axis]],
                                                [v0[axis - 1], v1[axis - 1]],
                                                kind="linear")
    for value in range(v0[axis], v1[axis], delta):
        if axis == 0:
            position = (value, interpolation(value))
        else:
            position = (interpolation(value), value)

        vertices_to_return.update(get_layer_elements(position, layers))
    return vertices_to_return


def arc_length(x, y):
    gradient = np.gradient(y, x)
    return np.trapz(np.sqrt(1 + gradient**2))


def modified_intensities(image):
    import scipy.optimize as scop
    # import matplotlib.pyplot as plt
    # import os

    # image = Image.open(image_path)
    # max_intensity = np.amax(image)
    # def linear_function(x, slope, angle, max_intensity):
    #     return max_intensity - slope * np.tan(angle) * x
    def linear_function(x, slope, A):
        return A + slope * x

    yrange = range(0, image.size[1])
    intensity_vs_y = [np.median(np.array(image)[jj, :]) for jj in yrange]

    # fit only the first and last point
    # parameters = scop.curve_fit(linear_function, yrange, intensity_vs_y)
    parameters = scop.curve_fit(linear_function,
                                [yrange[0], yrange[-1]],
                                [intensity_vs_y[0], intensity_vs_y[-1]])

    ####################################
    # modify the myosin intensities now
    ####################################
    corrected_pixels = np.array(image).copy().astype("int32")
    # for ii in range(0, image.size[0]):
    for jj in range(0, image.size[1]):
        corrected_pixels[jj, :] = corrected_pixels[jj, :] - linear_function(jj, parameters[0][0], parameters[0][1])

    # intensity_vs_y_post = [np.median(corrected_pixels[jj, :]) for jj in yrange]
    # _, axs = plt.subplots(2, 1, sharex=True)
    # effective_slope = round(parameters[0][0], 4)
    # a_value = round(parameters[0][1], 4)
    # axs[0].plot(yrange, intensity_vs_y)
    # axs[0].plot(yrange, linear_function(yrange, *parameters[0]), ls="--", color="red")
    # axs[0].text(0.75, 0.85, f"y = {a_value} + {effective_slope} * x",transform=axs[0].transAxes)
    # axs[0].set_ylabel("Intensity")

    # axs[1].plot(yrange, intensity_vs_y_post)
    # axs[1].set_ylabel("Intensity")

    # plt.xlabel("Y-Axis [px]")

    # plt.savefig(os.path.join("results_fit_test", image_path.split("/")[-1].split(".")[0]), dpi=350)
    # plt.close()

    return corrected_pixels



from dataclasses import dataclass, field

@dataclass 
class Vertex:
    """
    Class representation of the vertices objects.

    :param id: Unique identifier of the vertex
    :type id: int
    :param x: x-position of the vertex
    :type x: float
    :param y: y-position of the vertex
    :type y: float
    :param ownEdges: List of edges that connect with this vertex
    :type ownEdges: list
    :param ownCells: List of cells that connect with this vertex
    :type ownCells: list
    :param own_big_edges: Unique identifier of the vertex
    :type own_big_edges: int
    """
    
    id: int
    x: float
    y: float
    ownEdges: list = field(default_factory=list)
    ownCells: list = field(default_factory=list)
    own_big_edges: list = field(default_factory=list)
    
    def get_coords(self) -> list:
        """
        Return an array with the [x, y] coordinates of the vertex

        :return: [x, y] position of the vertex
        :rtype: list
        """
        return [self.x, self.y]

    def add_cell(self, cid: int) -> bool:
        """
        Add a new cell to the list of cells that include this vertex

        :param cid: ID of the cell to be added
        :type cid: int
        :return: False if the provided ID is already in the list
        :rtype: bool
        """

        if cid in self.ownCells:
            return False
        else:
            self.ownCells.append(cid)
            return True
        
    def remove_cell(self, cid: int) -> list:
        """
        Remove a given cell from the list of cells that include this vertex

        :param cid: ID of the cell to be removed
        :type cid: int
        :return: List of cells after removal
        :rtype: list
        """

        self.ownCells.remove(cid)
        return self.ownCells

    def add_edge(self, eid: int) -> bool:
        """
        Add a given id to the list of edges that include this vertex

        :param eid: ID of the edge to be added
        :type eid: int
        :return: False if the provided ID is already in the list
        :rtype: bool
        """

        if eid in self.ownEdges:
            print(eid, self.ownEdges)
            print("edge already in vertex")
            return False
        else:
            self.ownEdges.append(eid)
            return True
    
    def remove_edge(self, eid: int) -> list:
        """
        Remove a given edge from the list of edges that include this vertex

        :param eid: ID of the edge to be removed
        :type cid: int
        :return: List of edges after removal
        :rtype: list
        """
        self.ownEdges.remove(eid)
        return self.ownEdges

    def add_big_edge(self, beid: int) -> bool:
        """
        Add a given big edge id to the list of big edges that include this vertex

        :param beid: ID of the big edge to be added
        :type beid: int
        :return: False if the provided ID is already in the list
        :rtype: bool
        """
        if beid in self.own_big_edges:
            print(beid, self.own_big_edges)
            print("edge already in vertex")
            return False
        else:
            self.own_big_edges.append(beid)
            return True
    
    def remove_big_edge(self, beid: int) -> list:
        """
        Remove a given big edge from the list of big edges that include this vertex

        :param beid: ID of the big edge to be removed
        :type cid: int
        :return: List of big edges after removal
        :rtype: list
        """
        self.own_big_edges.remove(beid)
        return self.own_big_edges


import numpy as np
import pandas as pd
import os

import forsys.vertex as fvertex
import forsys.edge as fedge
import forsys.cell as fcell

def create_lattice(wkt):
    """
    Creates the lattice, give a wkt formated file, and returns
    vertex edges and cells as dictionaries to input into Lattice.
    """
    vertices = {}
    edges = {}
    cells = {}
    verticesNumber = 0
    edgesNumber = 0
    cellsNumber = 0
    edArr = []
    for cellNonParsed in wkt:
        # print("Creating cell ", cellsNumber)
        cellVertices = []
        arrPosition = []
        cell = read_vertexes(cellNonParsed)
        for vID in range(0, len(cell)-1):
            v = cell[vID].split(' ')
            # Check if the first vertex exists, if doesn't exist, create it
            oldVertex = is_vertex_created(v, vertices)
            if type(oldVertex) != int :  # Doesn't exist
                # vertices[verticesNumber] = fvertex.Vertex(verticesNumber, float(v[1]), float(v[2]))
                vertices[verticesNumber] = fvertex.Vertex(verticesNumber, float(v[1]), 1024 - float(v[2]))
                arrPosition.append(verticesNumber)
                verticesNumber += 1
            else:  # Save the global ID
                arrPosition.append(oldVertex)
        
        # vertexCellID = 0
        for v in range(0, len(arrPosition)):
            i = arrPosition[v]
            try:
                j = arrPosition[v+1]
            except IndexError:
                j = arrPosition[0]

            # cellVertices[vertexCellID] = vertices[i]
            cellVertices.append(vertices[i])
            # vertexCellID += 1
            if not (i, j) in edArr and not (j, i) in edArr:
                # edges[edgesNumber] = e.Edge(edgesNumber, vertices[i], vertices[j])
                edges[edgesNumber] = fedge.SmallEdge(edgesNumber, vertices[i], vertices[j])
                edArr.append((i, j))
                edgesNumber += 1
        # cellType = {'typeID':0,'targetArea': 600,'targetPerimeter':100,'areaElasticityConstant':1/2,'perimeterElasticityConstant':0.5/2,'cellHeight':1}
        cellType = {}
        cells[cellsNumber] = fcell.Cell(cellsNumber, cellVertices, cellType)
        cellsNumber += 1

    # vertices, edges, cells = reduce_amount(vertices, edges, cells)

    return vertices, edges, cells


def reduce_amount(vertices, edges, cells):
    for _, cell in cells.items():
        for j in cell.vertices:
            i = cell.get_next_vertex(j)
            # j = nv
            k = cell.get_previous_vertex(j)

            ipos = i.get_coords()
            jpos = j.get_coords()
            kpos = k.get_coords()

            deltaX1 = jpos[0]-ipos[0]
            deltaY1 = jpos[1]-ipos[1]
            deltaX2 = kpos[0]-jpos[0]
            deltaY2 = kpos[1]-jpos[1]

            if deltaX1 != 0:
                slope1 = deltaY1/deltaX1
            else:
                slope1 = np.inf
            if deltaX2 != 0:
                slope2 = deltaY2/deltaX2
            else:
                slope2 = np.inf
                
            if slope1 == slope2:
                toRemove = None
                toReplace = None

                if len(j.ownCells) < 3 and len(j.ownEdges) < 3:
                    for e in j.ownEdges:
                        if i.id in edges[e].get_vertices_id():
                            toReplace = e
                        elif k.id in edges[e].get_vertices_id():
                            toRemove = e
                    # edges[toRemove].v1.ownEdges.remove(toRemove)
                    # edges[toRemove].v2.ownEdges.remove(toRemove)
                    del edges[toRemove]
                    edges[toReplace].replace_vertex(j, k)
                    for c in j.ownCells:
                        cells[c].vertices.remove(j)
                    del vertices[j.id]

    return vertices, edges, cells


def read_vertexes(row):
    var = row.split('((')[1].split(',')
    var[0] = ' '+var[0]
    var[-1] = var[-1][:-2]
    # var = var[:-1] # Why is this line necessary ??
    return var


def is_vertex_created(v, vertices):
    # Check all vertices to see if we already have one in that position
    for _, vertex in vertices.items():
        if float(v[1]) == float(vertex.x) and 1024 - float(v[2]) == float(vertex.y):
            return vertex.id
    return False

def create_wkt(cells):
    """
    Creates a WKT-formated string from a lattice.
    POLYGON ((x0 y0, x1 y1, x2 y2))
    go through each vertex in a cell and save the position with round(*, 3)
    """
    final = str()
    for _, cell in cells.items():
        string = str()
        for v in cell.vertices:
          string += f"{str(v.x)} {str(v.y)}, "
        string = string[:-2]
        final += "POLYGON ((" + string+ "))"+"\n"

    return final

def save(wkt, folder="log/",step=0):
    """ Save the current state """
    with open(os.path.join(folder, str(step)+".wkt"), 'w') as f:
        f.write(wkt)
    # return wkt.to_csv(folder+"wkt-"+str(step)+".wkt", index=False, header=None)



from forsys.forsys import ForSys

from . import surface_evolver
from . import frames
from . import plot
from . import virtual_edges
from . import auxiliar
from . import wkt
from . import skeleton
from . import exceptions
from . import myosin
from . import stress_tensor
from . import tessellation
# from . import general_matrix
import numpy as np
np.seterr(all='raise')
